import json
from typing import TYPE_CHECKING, Any, Literal, overload

from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ssm.literals import ParameterTypeType
else:
    ParameterTypeType = str

get_ssm_client = AWSService.SSM.get_client


def put_ssm_parameter(
    param_name: str,
    param_value: str,
    param_type: ParameterTypeType = "String",
    exists_ok: bool = True,
) -> int:
    """Put or update parameter

    Args:
        param_name (str): Name of parameter
        param_value (str): New value of parameter
        param_type (ParameterTypeType, optional): Type of parameter. Defaults to "String".
        exists_ok (bool, optional): Allow overwrite if value exists. Defaults to True.

    Returns:
        The version of the parameter
    """
    ssm = get_ssm_client()
    return ssm.put_parameter(
        Name=param_name, Value=param_value, Type=param_type, Overwrite=exists_ok
    )["Version"]


@overload
def get_ssm_parameter(
    param_name: str, as_dict: Literal[True]
) -> dict[str, Any]: ...  # pragma: no cover


@overload
def get_ssm_parameter(
    param_name: str, as_dict: Literal[False] = False
) -> str: ...  # pragma: no cover


def get_ssm_parameter(param_name: str, as_dict: bool = False) -> str | dict[str, Any]:
    """Retrieves a SSM parameter value

    Args:
        param_name (str): the SSM parameter key name
        as_dict (bool, optional): Return as dict. Defaults to False.

    Raises:
        ValueError: If there is no such key

    Returns:
        The parameter value stored at the key name
    """
    ssm = get_ssm_client()

    response = ssm.get_parameter(Name=param_name, WithDecryption=True)

    param = response["Parameter"]

    param_value = param.get("Value", None)

    if param_value is None:
        raise ValueError(f"Error obtaining param {param_name} from parameter store.")

    if as_dict:
        return json.loads(param_value)
    return param_value


def has_ssm_parameter(param_name: str) -> bool:
    """Check if an SSM parameter exists.

    Args:
        param_name (str): The SSM parameter key name to check.

    Returns:
        True if the parameter exists, False otherwise.
    """
    try:
        get_ssm_parameter(param_name)
    except (ValueError, ClientError):
        return False
    else:
        return True



# Standard Environment variables from AWS Lambda:
# https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html#configuration-envvars-runtime
AWS_LAMBDA_FUNCTION_NAME_KEY = "AWS_LAMBDA_FUNCTION_NAME"
AWS_LAMBDA_FUNCTION_VERSION_KEY = "AWS_LAMBDA_FUNCTION_VERSION"
AWS_LAMBDA_FUNCTION_MEMORY_SIZE_KEY = "AWS_LAMBDA_FUNCTION_MEMORY_SIZE"
AWS_LAMBDA_LOG_GROUP_NAME_KEY = "AWS_LAMBDA_LOG_GROUP_NAME"
AWS_LAMBDA_LOG_STREAM_NAME_KEY = "AWS_LAMBDA_LOG_STREAM_NAME"
AWS_EXECUTION_ENV_KEY = "AWS_EXECUTION_ENV"
AWS_LAMBDA_FUNCTION_HANDLER_STANDARD_KEY = "_HANDLER"

# These are custom env variable names
AWS_LAMBDA_FUNCTION_REQUEST_ID_KEY = "AWS_LAMBDA_FUNCTION_REQUEST_ID"
AWS_LAMBDA_FUNCTION_ARN_KEY = "AWS_LAMBDA_FUNCTION_ARN"
AWS_LAMBDA_FUNCTION_HANDLER_KEY = "AWS_LAMBDA_FUNCTION_HANDLER"
AWS_LAMBDA_EVENT_PAYLOAD_KEY = "AWS_LAMBDA_EVENT_PAYLOAD"
AWS_LAMBDA_EVENT_RESPONSE_LOCATION_KEY = "AWS_LAMBDA_EVENT_RESPONSE_LOCATION"


DEFAULT_AWS_LAMBDA_FUNCTION_NAME = "unknown"



import json
import logging

from aibs_informatics_core.utils.json import DecimalEncoder
from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService, get_region
from aibs_informatics_aws_utils.exceptions import AWSError

logger = logging.getLogger(__name__)


get_sqs_client = AWSService.SQS.get_client
get_sqs_resource = AWSService.SQS.get_resource


def delete_from_queue(queue_name: str, receipt_handle: str, region: str | None = None):
    """Delete a message from an SQS queue.

    Args:
        queue_name (str): The name of the SQS queue.
        receipt_handle (str): The receipt handle of the message to delete.
        region (Optional[str]): AWS region. Defaults to None (uses default region).

    Returns:
        True if the message was successfully deleted.
    """
    sqs = get_sqs_client()
    queue_url_response = sqs.get_queue_url(QueueName=queue_name)
    queue_url = queue_url_response["QueueUrl"]

    sqs.delete_message(QueueUrl=queue_url, ReceiptHandle=receipt_handle)
    logger.info("Deleted message %s form queue %s", receipt_handle, queue_name)

    return True


def send_to_dispatch_queue(payload: dict, env_base: str):
    """Send a message to the demand request dispatch queue.

    Args:
        payload (dict): The message payload as a dictionary.
        env_base (str): The environment base used to construct the queue name.

    Raises:
        AWSError: If the queue cannot be found.

    Returns:
        The MD5 digest of the sent message body.
    """
    sqs = get_sqs_client(region=get_region())
    queue_name = "-".join([env_base, "demand_request_queue"])
    logger.info("Queue name: %s", queue_name)

    try:
        queue_url_response = sqs.get_queue_url(QueueName=queue_name)
    except ClientError as e:
        logger.exception(e)
        raise AWSError(f"Could not find SQS queue with name {queue_name}")

    queue_url = queue_url_response["QueueUrl"]

    response = sqs.send_message(QueueUrl=queue_url, MessageBody=json.dumps(payload))

    return response["MD5OfMessageBody"]


def send_sqs_message(
    queue_name: str,
    payload: dict,
    message_deduplication_id: str | None = None,
    message_group_id: str | None = None,
    payload_json_encoder: type[json.JSONEncoder] = DecimalEncoder,
) -> str:
    """Send a message to an SQS queue by providing a queue name

    Args:
        queue_name (str): The name of the queue that you want to send a message to.
            (e.g. 'aps-sync-request-queue.fifo')
        payload (dict): A dictionary representing the message payload you would like to send.
        message_deduplication_id (Optional[str], optional): An ID that can be used by SQS
            to remove messages that have the same deduplication_id. Do not set if your
            SQS queue already uses content based deduplication. Defaults to None.
        message_group_id (Optional[str], optional): Required for FIFO queues.
            Messages sent with the same message_group_id will obey FIFO rules. Messages with
            different message_group_ids may be interleaved. Defaults to None.
        payload_json_encoder (Type[json.JSONEncoder], optional): The JSONEncoder
            class that should be used to covert the input `payload` dictionary into
            a json string. By default uses a DecimalEncoder which can handle decimal.Decimal types.

    Raises:
        AWSError: If the provided queue_name cannot be resolved to an SQS url.
            HINT: Does the code calling this function have the correct SQS permissions?
        RuntimeError: If the destination queue is a FIFO queue, then `message_group_id` MUST
            be provided.

    Returns:
        Returns an MD5 digest of the send message body.
    """
    sqs = get_sqs_client(region=get_region())
    try:
        queue_url_response = sqs.get_queue_url(QueueName=queue_name)
    except ClientError:
        raise AWSError(
            f"Could not find SQS queue with name: {queue_name}. "
            "Does the code calling send_sqs_message() have sqs:GetQueueUrl permissions?"
        )

    send_sqs_message_args = {
        "QueueUrl": queue_url_response["QueueUrl"],
        "MessageBody": json.dumps(payload, cls=payload_json_encoder),
    }

    if message_group_id is not None:
        send_sqs_message_args["MessageGroupId"] = message_group_id
    else:
        if queue_name.endswith(".fifo"):
            raise RuntimeError("SQS messages for a FIFO queue *must* include a message_group_id!")

    if message_deduplication_id is not None:
        send_sqs_message_args["MessageDeduplicationId"] = message_deduplication_id

    response = sqs.send_message(**send_sqs_message_args)  # type: ignore  # complains about valid kwargs

    return response["MD5OfMessageBody"]



from __future__ import annotations

__all__ = ["BaseFileSystem", "LocalFileSystem", "S3FileSystem"]

import errno
import os
from abc import abstractmethod
from collections import deque
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

import pytz
from aibs_informatics_core.models.aws.efs import EFSPath
from aibs_informatics_core.models.aws.s3 import S3Path
from aibs_informatics_core.models.base import AwareIsoDateTime, PydanticBaseModel
from aibs_informatics_core.utils.logging import get_logger
from aibs_informatics_core.utils.os_operations import find_all_paths
from aibs_informatics_core.utils.time import BEGINNING_OF_TIME
from aibs_informatics_core.utils.tools.strtools import removeprefix

from aibs_informatics_aws_utils.efs import get_efs_path, get_local_path
from aibs_informatics_aws_utils.s3 import get_s3_resource

logger = get_logger(__name__)

SEP = "/"


class PathStats(PydanticBaseModel):
    size_bytes: int
    object_count: int
    last_modified: AwareIsoDateTime


@dataclass(order=True)
class Node:
    """Represents an object or folder in an file system path.

    Attributes:
        path_part: Specifies the key part of the fs path (an edge) to this node.
        parent: Optionally specify the parent node to which
            this node is connected. By default, this is None.
        children: Child nodes that exist under this path prefix.
        size_bytes: The size (in bytes) of all objects under this path prefix.
        object_count: The number of objects under this path prefix.
        last_modified: The most recent date any objects under this prefix were
            last modified.
    """

    path_part: str
    parent: Node | None = field(default=None)
    children: dict[str, Node] = field(default_factory=dict)
    size_bytes: int = field(default=0)
    object_count: int = field(default=0)
    last_modified: datetime = field(default=BEGINNING_OF_TIME)
    is_path_part_prefix: bool = field(default=False)
    is_path_part_suffix: bool = field(default=False)

    def __hash__(self) -> int:
        return hash(self.path)

    @property
    def key(self) -> str:
        return self.path

    @property
    def parent_path(self) -> str:
        parent_path = self.parent.path if self.parent else ""
        if parent_path and self.is_path_part_suffix:
            parent_path = parent_path.rstrip(SEP)
        return parent_path

    @property
    def normalized_path_part(self) -> str:
        path_part = self.path_part
        if self.has_children() and not self.is_path_part_prefix:
            path_part = path_part.rstrip(SEP) + SEP
        return path_part

    @property
    def path(self) -> str:
        return self.parent_path + self.normalized_path_part

    @property
    def path_stats(self) -> PathStats:
        return PathStats(
            size_bytes=self.size_bytes,
            object_count=self.object_count,
            last_modified=self.last_modified,
        )

    @property
    def depth(self) -> int:
        return self.parent.depth + 1 if self.parent else 0

    def has_children(self) -> bool:
        return not (len(self.children) < 1)

    def add_object(self, path: str, size: int, last_modified: datetime):
        def _add_object(node: Node, path: str | None):
            node._update_stats(size=size, last_modified=last_modified)
            if path is None:
                return

            first_key_part, remaining_key = path.split(SEP, 1) if SEP in path else (path, None)
            if first_key_part:
                if first_key_part not in node.children:
                    node.children[first_key_part] = Node(path_part=first_key_part, parent=node)
                node = node.children[first_key_part]
            _add_object(node, remaining_key)

        # TODO: Right now, we cannot support non-folder prefixes
        _add_object(self, path.lstrip(SEP))

    def get(self, key: str) -> Node | None:
        try:
            return self[key]
        except KeyError:
            return None

    def list_nodes(self) -> list[Node]:
        nodes = [self]
        for _, n in self.children.items():
            nodes.extend(n.list_nodes())
        return nodes

    def _update_stats(self, size: int, last_modified: datetime):
        # For each node, update the current node's stats
        self.size_bytes += size
        self.object_count += 1
        if self.last_modified < last_modified:
            self.last_modified = last_modified

    def __getitem__(self, key: str) -> Node:
        _self = self
        for key_part in key.split(SEP):
            # Only access if value is not empty string
            if key_part:
                _self = _self.children[key_part]
        return _self

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"path_part={self.path_part}, "
            f"children={len(self.children)}, "
            f"size_bytes={self.size_bytes}, "
            f"object_count={self.object_count})"
        )


@dataclass  # type: ignore[misc] # mypy #5374
class BaseFileSystem:
    node: Node = field(init=False)

    def __post_init__(self):
        self.node = self.initialize_node()

    @abstractmethod
    def initialize_node(self) -> Node:
        raise NotImplementedError()

    @abstractmethod
    def refresh(self, **kwargs):
        raise NotImplementedError()

    def partition(
        self,
        size_bytes_limit: int | None = None,
        object_count_limit: int | None = None,
        raise_error_if_criteria_not_met: bool = False,
    ) -> list[Node]:
        """Partitions the root tree folder structure into a list of nodes.

        Partitioning is guided by constraints by size and object count.

        Args:
            size_bytes_limit: If specified, partitions must be less than the specified value.
            object_count_limit: If specified, partitions must contain fewer objects than
                the specified value.
            raise_error_if_criteria_not_met: If True, raises error if nodes cannot meet
                criteria. In actuality, this is more relevant for size limitations where
                an object size is greater than the size limit.

        Raises:
            ValueError: Thrown if raise_error_if_criteria_not_met is true and criteria not met.

        Returns:
            List of nodes representing the partition.
        """
        unchecked_nodes = {self.node}
        size_bytes_exceeding_obj_nodes = []

        partitioned_nodes: list[Node] = []
        logger.info(
            f"Partitioning nodes with size_bytes_limit={size_bytes_limit} "
            f"and object_count_limit={object_count_limit}"
        )

        while unchecked_nodes:
            unchecked_node = unchecked_nodes.pop()
            if (size_bytes_limit and unchecked_node.size_bytes > size_bytes_limit) or (
                object_count_limit and unchecked_node.object_count > object_count_limit
            ):
                if unchecked_node.has_children():
                    unchecked_nodes.update(unchecked_node.children.values())
                else:
                    size_bytes_exceeding_obj_nodes.append(unchecked_node)
            else:
                partitioned_nodes.append(unchecked_node)

        if size_bytes_exceeding_obj_nodes:
            msg = (
                f"Found {len(size_bytes_exceeding_obj_nodes)} objects that exceed the "
                f"partition size limit {size_bytes_limit}."
            )
            if raise_error_if_criteria_not_met:
                raise ValueError(msg)
            logger.warning(msg)
            partitioned_nodes.extend(size_bytes_exceeding_obj_nodes)
        logger.info(f"Partitioned {len(partitioned_nodes)} nodes.")
        return partitioned_nodes

    @classmethod
    @abstractmethod
    def from_path(cls, path: str, **kwargs) -> BaseFileSystem:
        pass


@dataclass
class LocalFileSystem(BaseFileSystem):
    path: Path

    def initialize_node(self) -> Node:
        return Node(path_part=self.path.as_posix())

    def refresh(self, **kwargs):
        self.node = self.initialize_node()
        paths_to_visit = deque(find_all_paths(self.path, include_dirs=False, include_files=True))
        while paths_to_visit:
            path = paths_to_visit.popleft()
            try:
                path_stats = Path(path).stat()
                self.node.add_object(
                    path=removeprefix(path, str(self.path) + os.sep),
                    size=path_stats.st_size,
                    last_modified=datetime.fromtimestamp(path_stats.st_mtime, tz=pytz.UTC),
                )
            except FileNotFoundError:
                logger.warning(f"{path} does not exist. Not adding to {self}")
            except OSError as ose:
                # Suppress error if Stale File. This is expected error if file has been deleted:
                #   - https://stackoverflow.com/a/40351967
                #   - https://www.rfc-editor.org/rfc/rfc7530#section-4
                if ose.errno == errno.ESTALE:
                    logger.warning(f"{ose} raised for {path}.")
                    if Path(path).exists():
                        logger.warning(f"Adding {path} to end of list to check later.")
                        paths_to_visit.append(path)
                else:
                    logger.error(f"Unexpected error raised for {path}. Reason: {ose}")
                    raise ose

    @classmethod
    def from_path(cls, path: str | Path, **kwargs) -> LocalFileSystem:
        local_path = Path(path)
        local_root = LocalFileSystem(path=local_path)
        local_root.refresh(**kwargs)
        return local_root


@dataclass
class EFSFileSystem(LocalFileSystem):
    efs_path: EFSPath

    def initialize_node(self) -> Node:
        return Node(path_part=self.efs_path)

    @classmethod
    def from_path(cls, path: str | Path, **kwargs) -> EFSFileSystem:
        if isinstance(path, str) and EFSPath.is_valid(path):
            efs_path = EFSPath(path)
            local_path = get_local_path(efs_path=efs_path)
        else:
            local_path = Path(path)
            efs_path = get_efs_path(local_path=local_path)

        efs_root = EFSFileSystem(path=local_path, efs_path=efs_path)
        efs_root.refresh(**kwargs)
        return efs_root


@dataclass
class S3FileSystem(BaseFileSystem):
    """Generates a FS tree structure of an S3 path with size and object count stats.

    Attributes:
        bucket: The S3 bucket to describe.
        key: The S3 key to describe.
    """

    bucket: str
    key: str

    def initialize_node(self) -> Node:
        return Node(path_part=self.key)

    def refresh(self, **kwargs):
        self.node = self.initialize_node()
        s3 = get_s3_resource(**kwargs)
        bucket = s3.Bucket(self.bucket)

        for obj in bucket.objects.filter(Prefix=self.key):
            self.node.add_object(
                path=removeprefix(obj.key, self.key),
                size=obj.size,
                last_modified=obj.last_modified,
            )

    @classmethod
    def from_path(cls, path: str, **kwargs) -> S3FileSystem:
        s3_path = S3Path(path)
        s3_root = S3FileSystem(bucket=s3_path.bucket, key=s3_path.key)
        s3_root.refresh(**kwargs)
        return s3_root


def get_file_system(path: str | Path) -> BaseFileSystem:
    if isinstance(path, str) and S3Path.is_valid(path):
        return S3FileSystem.from_path(path)
    elif isinstance(path, str) and EFSPath.is_valid(path):
        return EFSFileSystem.from_path(path)
    else:
        return LocalFileSystem.from_path(path)



import urllib.parse

from aibs_informatics_core.utils.logging import get_logger

from aibs_informatics_aws_utils.core import AWSService, get_region

logger = get_logger(__name__)


get_logs_client = AWSService.LOGS.get_client


def build_log_stream_url(
    log_group_name: str,
    log_stream_name: str,
    region: str | None = None,
) -> str:
    """Build a CloudWatch Logs console URL for a log stream.

    Args:
        log_group_name (str): The name of the CloudWatch log group.
        log_stream_name (str): The name of the log stream within the group.
        region (Optional[str]): AWS region. Defaults to None (uses default region).

    Returns:
        The CloudWatch console URL for viewing the log stream.
    """

    def _special_escape(string: str) -> str:
        return urllib.parse.quote(string, safe="").replace("%", "$25")

    region = get_region(region)
    log_group_name = _special_escape(log_group_name)
    log_stream_name = _special_escape(log_stream_name)

    log_url_prefix = f"https://{region}.console.aws.amazon.com/cloudwatch/home?region={region}#logsV2:log-groups/log-group/"
    return f"{log_url_prefix}{log_group_name}/log-events/{log_stream_name}"



import logging
import time
from typing import TYPE_CHECKING, Literal, Unpack

from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import AWSService
from aibs_informatics_aws_utils.exceptions import AWSError

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_athena.type_defs import (
        GetQueryExecutionOutputTypeDef,
        QueryExecutionStatusTypeDef,
        QueryExecutionTypeDef,
        StartQueryExecutionInputTypeDef,
        StartQueryExecutionOutputTypeDef,
    )
else:
    GetQueryExecutionOutputTypeDef = dict
    QueryExecutionStatusTypeDef = dict
    QueryExecutionTypeDef = dict
    StartQueryExecutionInputTypeDef = dict
    StartQueryExecutionOutputTypeDef = dict


ATHENA_QUERY_WAITER_STATUS = Literal["SUCCEEDED", "FAILED", "CANCELLED", "TIMEOUT"]

logger = logging.getLogger(__name__)

get_athena_client = AWSService.ATHENA.get_client


def start_query_execution(
    query_string: str,
    work_group: str | None = None,
    execution_parameters: list[str] | None = None,
    **kwargs: Unpack[StartQueryExecutionInputTypeDef],
) -> StartQueryExecutionOutputTypeDef:
    """Start an Athena query execution.

    Args:
        query_string (str): The SQL query string to execute.
        work_group (Optional[str]): The name of the workgroup to execute the query in.
        execution_parameters (Optional[List[str]]): Optional list of query execution parameters.
        **kwargs: Additional arguments passed to the Athena start_query_execution API.

    Raises:
        AWSError: If the query execution fails to start.

    Returns:
        The start query execution response containing the QueryExecutionId.
    """
    athena = get_athena_client()

    request = StartQueryExecutionInputTypeDef(QueryString=query_string)
    if work_group:
        request["WorkGroup"] = work_group
    if execution_parameters:
        request["ExecutionParameters"] = execution_parameters
    request.update(kwargs)
    try:
        metadata = athena.start_query_execution(**request)
        return metadata
    except ClientError as e:
        logger.error(f"Error executing : {request} {e}", exc_info=True)
        raise AWSError(f"Error starting query execution: {request} {e}") from e


def get_query_execution(query_execution_id: str) -> GetQueryExecutionOutputTypeDef:
    """Get the status and details of an Athena query execution.

    Args:
        query_execution_id (str): The unique identifier of the query execution.

    Raises:
        AWSError: If the query execution cannot be retrieved.

    Returns:
        The query execution details including status and results location.
    """
    athena = get_athena_client()
    try:
        return athena.get_query_execution(QueryExecutionId=query_execution_id)
    except Exception as e:
        logger.error(f"Error executing : {query_execution_id} {e}", exc_info=True)
        raise AWSError(f"Error starting query execution: {query_execution_id} {e}") from e


def query_waiter(
    query_execution_id: str, timeout: int = 60
) -> tuple[ATHENA_QUERY_WAITER_STATUS, QueryExecutionStatusTypeDef]:
    """Wait for an Athena query to complete.

    Polls the query execution status until it reaches a terminal state
    (SUCCEEDED, FAILED, CANCELLED) or times out.

    Args:
        query_execution_id (str): The unique identifier of the query execution.
        timeout (int): Maximum time to wait in seconds. Defaults to 60.

    Returns:
        A tuple of (status, status_details) where status is one of
            SUCCEEDED, FAILED, CANCELLED, or TIMEOUT.
    """
    start = time.time()
    logger.info(f"Polling for status of query execution: {query_execution_id}")
    while True:
        stats = get_query_execution(query_execution_id=query_execution_id)
        logger.info(f"Query Execution Status: {stats}")
        status = stats["QueryExecution"].get("Status", {})
        state = status.get("State")
        if state in ["SUCCEEDED", "FAILED", "CANCELLED", "TIMEOUT"]:
            return state, status  # type: ignore[return-value]
        time.sleep(0.2)  # 200ms
        # Exit if the time waiting exceed the timeout seconds
        if time.time() > start + timeout:
            return "TIMEOUT", status



from __future__ import annotations

__all__ = [
    "get_local_path",
]

import logging
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Union, overload

from aibs_informatics_core.models.aws.efs import EFSPath

from aibs_informatics_aws_utils.efs.mount_point import MountPointConfiguration, detect_mount_points

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_efs.type_defs import (
        AccessPointDescriptionTypeDef,
        DescribeAccessPointsRequestPaginateTypeDef,
        DescribeAccessPointsResponseTypeDef,
        DescribeFileSystemsResponseTypeDef,
        FileSystemDescriptionTypeDef,
        TagTypeDef,
    )
else:
    AccessPointDescriptionTypeDef = dict
    DescribeFileSystemsResponseTypeDef = dict
    DescribeAccessPointsRequestPaginateTypeDef = dict
    DescribeAccessPointsResponseTypeDef = dict
    FileSystemDescriptionTypeDef = dict
    TagTypeDef = dict

logger = logging.getLogger(__name__)

StrPath = Union[Path, str]


@overload
def get_efs_path(
    local_path: Path,
    raise_if_unresolved: Literal[False],
    mount_points: list[MountPointConfiguration] | None = None,
) -> EFSPath | None: ...


@overload
def get_efs_path(
    local_path: Path,
    raise_if_unresolved: Literal[True] = True,
    mount_points: list[MountPointConfiguration] | None = None,
) -> EFSPath: ...


def get_efs_path(
    local_path: Path,
    raise_if_unresolved: bool = True,
    mount_points: list[MountPointConfiguration] | None = None,
) -> EFSPath | None:
    """Converts a local path assumed to be on a mount point to the EFS path

    Args:
        local_path (Path): Local path
        raise_if_unresolved (bool): If True, raises an error if the local path is not
            under an identifiable mount point. Defaults to True.
        mount_points (List[MountPointConfiguration] | None): Optionally can override
            list of mount_points. If None, mount points are detected. Defaults to None.

    Returns:
        Corresponding EFS URI or None if the path cannot be resolved and
            raise_if_unresolved is False
    """
    mount_points = mount_points if mount_points is not None else detect_mount_points()

    for mp in mount_points:
        if mp.is_mounted_path(local_path):
            logger.debug(f"Found mount point {mp} that matches path {local_path}")
            return mp.as_efs_uri(local_path)
    else:
        message = (
            f"Local path {local_path} is not relative to any of the "
            f"{len(mount_points)} mount point mount_points. Adapters: {mount_points}"
        )
        if raise_if_unresolved:
            logger.error(message)
            raise ValueError(message)
        logger.warning(message)
        return None


@overload
def get_local_path(
    efs_path: EFSPath,
    raise_if_unmounted: Literal[False],
    mount_points: list[MountPointConfiguration] | None = None,
) -> Path | None: ...


@overload
def get_local_path(
    efs_path: EFSPath,
    raise_if_unmounted: Literal[True] = True,
    mount_points: list[MountPointConfiguration] | None = None,
) -> Path: ...


def get_local_path(
    efs_path: EFSPath,
    raise_if_unmounted: bool = True,
    mount_points: list[MountPointConfiguration] | None = None,
) -> Path | None:
    """Gets a valid locally mounted path for the given EFS path.

    Args:
        efs_path (EFSPath): The EFS path. e.g., "efs://fs-12345678:/path/to/file.txt"
        raise_if_unmounted (bool): If True, raises an error if the EFS path is
            not mounted locally. Defaults to True.
        mount_points (List[MountPointConfiguration] | None): Optionally can override
            list of mount points. If None, mount points are detected. Defaults to None.

    Returns:
        The local path. e.g., "/mnt/efs/path/to/file.txt" or None if the path
            cannot be resolved and raise_if_unmounted is False
    """
    mount_points = mount_points if mount_points is not None else detect_mount_points()
    for mount_point in mount_points:
        if mount_point.file_system["FileSystemId"] == efs_path.file_system_id:
            logger.debug(
                f"Found {mount_point} with matching file system id for efs path {efs_path}"
            )

            if not efs_path.path.is_relative_to(mount_point.access_point_path):
                logger.debug(
                    f"EFS Path {efs_path.path} is not relative "
                    f"to mount point access point {mount_point.access_point_path}. Skipping"
                )
                continue
            logger.info(f"Found matching mount point {mount_point} for efs path {efs_path}")
            return mount_point.as_mounted_path(efs_path.path)
    else:
        message = (
            f"Could not resolve local path for EFS path {efs_path} from "
            f"{len(mount_points)} mount points detected on host."
        )
        if raise_if_unmounted:
            logger.error(message)
            raise ValueError(message)
        logger.warning(message)
        return None



import json
from typing import Any, Literal, overload

from aibs_informatics_aws_utils.core import AWSService

get_secretsmanager_client = AWSService.SECRETSMANAGER.get_client


@overload
def get_secret_value(
    secret_name: str, as_dict: Literal[False] = False, region: str | None = None
) -> str: ...


@overload
def get_secret_value(
    secret_name: str, as_dict: Literal[True], region: str | None = None
) -> dict[str, Any]: ...


def get_secret_value(
    secret_name: str, as_dict: bool = False, region: str | None = None
) -> str | dict[str, Any]:
    """Retrieves a Secrets Manager secret value

    Args:
        secret_name (str): the Secrets Manager secret name
        as_dict (bool): If True, return the secret as a dictionary. Defaults to False.
        region (Optional[str]): AWS region. Defaults to None.

    Raises:
        ValueError: If there is no such key

    Returns:
        Union[str, dict]: The secret value stored at the key name, as string or dict
    """
    secretsmanager = get_secretsmanager_client(region=region)

    response = secretsmanager.get_secret_value(SecretId=secret_name)

    secret = response["SecretString"]
    if as_dict:
        return json.loads(secret)
    else:
        return secret



import functools
import warnings
from collections.abc import Mapping, MutableMapping, Sequence
from dataclasses import dataclass, field
from typing import (
    Any,
    Generic,
    Literal,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)

from aibs_informatics_core.env import EnvBase
from aibs_informatics_core.exceptions import ValidationError
from aibs_informatics_core.models.db import (
    DBIndex,
    DBModel,
    DynamoDBItemValue,
    DynamoDBKey,
    DynamoDBPrimaryKeyItemValue,
)
from aibs_informatics_core.utils.logging import LoggingMixin
from boto3.dynamodb.conditions import (
    Attr,
    BeginsWith,
    Between,
    BuiltConditionExpression,
    ConditionBase,
    ConditionExpressionBuilder,
    Equals,
    GreaterThan,
    GreaterThanEquals,
    Key,
    LessThan,
    LessThanEquals,
)
from botocore.exceptions import ClientError

from aibs_informatics_aws_utils.core import get_client_error_code
from aibs_informatics_aws_utils.dynamodb.conditions import condition_to_str
from aibs_informatics_aws_utils.dynamodb.functions import (
    convert_floats_to_decimals,
    execute_partiql_statement,
    table_delete_item,
    table_get_item,
    table_get_items,
    table_put_item,
    table_query,
    table_scan,
    table_update_item,
)
from aibs_informatics_aws_utils.exceptions import (
    DBQueryException,
    DBReadException,
    DBWriteException,
    EmptyQueryResultException,
    NonUniqueQueryResultException,
)

DB_TABLE = TypeVar("DB_TABLE", bound="DynamoDBTable")
DB_MODEL = TypeVar("DB_MODEL", bound=DBModel)
DB_INDEX = TypeVar("DB_INDEX", bound=DBIndex)


def check_db_query_unique(
    index: DBIndex | None,
    query_result: list[dict[str, Any]],
    key_condition_expression: ConditionBase | None = None,
    filter_expression: ConditionBase | None = None,
):
    if len(query_result) > 1:
        readable_key_expression: BuiltConditionExpression | None = None
        if key_condition_expression:
            expression_builder = ConditionExpressionBuilder()
            readable_key_expression = expression_builder.build_expression(
                condition=key_condition_expression, is_key_condition=True
            )
        raise NonUniqueQueryResultException(
            f"Querying '{index.table_name() if index else ''}' table "
            f"(index: {index.index_name if index else ''}, "
            f"key condition: {readable_key_expression}, "
            f"filters: {filter_expression}) "
            f"did not return EXACTLY 1 result! Query results: '{query_result}'"
        )


def check_db_query_non_empty(
    index: DBIndex | None,
    query_result: list[dict[str, Any]],
    key_condition_expression: ConditionBase | None = None,
    filter_expression: ConditionBase | None = None,
):
    if len(query_result) == 0:
        readable_key_expression: BuiltConditionExpression | None = None
        if key_condition_expression:
            expression_builder = ConditionExpressionBuilder()
            readable_key_expression = expression_builder.build_expression(
                condition=key_condition_expression, is_key_condition=True
            )
        raise EmptyQueryResultException(
            f"Querying '{index.table_name() if index else ''}' table "
            f"(index: {index.index_name if index else ''}, "
            f"key condition: {readable_key_expression}, "
            f"filters: {filter_expression}) "
            f"returned NO results when at least 1 result was expected!"
        )


def check_table_name_and_index_match(table_name: str, index: DBIndex | type[DBIndex]):
    if not table_name.endswith(index.table_name()):
        raise DBQueryException(
            f"The provided DBIndex ({index}) is not valid for the table to be queried "
            f"({table_name})!"
        )


def check_index_supports_strongly_consistent_read(index: DBIndex):
    if not index.supports_strongly_consistent_read:
        raise DBQueryException(
            f"The provided DBIndex ({index}) is a GSI/LSI of the table "
            f"({index.table_name()}) and does not support strongly consistent reads!"
        )


def build_optimized_condition_expression_set(
    candidate_indexes: type[DB_INDEX] | Sequence[DB_INDEX],
    *args: DynamoDBKey | ConditionBase,
    **kwargs: Any,
) -> tuple[DB_INDEX | None, ConditionBase | None, ConditionBase | None, list[ConditionBase]]:
    """Builds an optimized set of conditions for a query or scan


    Args:
        candidate_indexes (Type[DB_INDEX]|Sequence[DB_INDEX]): index class or subset of indexes
            the order of the indexes matters! The first index that matches the provided
            conditions will be used.
        *args (Union[DynamoDBKey, ConditionBase]): varargs of DynamoDBKey or ConditionBase
        **kwargs (Any): kwargs of DynamoDBKey or ConditionBase

    Returns:
        A tuple containing:
            the target index,
            partition key condition,
            sort key condition,
            and filter expressions.

    """
    target_index: DB_INDEX | None = None
    partition_key: ConditionBase | None = None
    sort_key_condition_expression: ConditionBase | None = None
    filter_expressions: list[ConditionBase] = []

    if not args and not kwargs:
        return target_index, partition_key, sort_key_condition_expression, filter_expressions

    if not isinstance(candidate_indexes, Sequence):
        candidate_indexes = [ci for ci in candidate_indexes]
    index_all_key_names = set(
        {
            *{_.key_name for _ in candidate_indexes},
            *{_.sort_key_name for _ in candidate_indexes if _.sort_key_name},
        }
    )

    SupportedKeyComparisonTypes = (
        Equals,
        GreaterThan,
        GreaterThanEquals,
        LessThan,
        LessThanEquals,
        BeginsWith,
        Between,
    )

    candidate_conditions: dict[
        str,
        (
            Equals
            | GreaterThan
            | GreaterThanEquals
            | LessThan
            | LessThanEquals
            | BeginsWith
            | Between
        ),
    ] = {}
    non_candidate_conditions: list[ConditionBase] = []

    for _ in (kwargs,) + args:
        if not isinstance(_, ConditionBase):
            for k, v in _.items():
                if k not in index_all_key_names:
                    non_candidate_conditions.append(Attr(k).eq(v))
                    continue
                new_condition = Key(k).eq(v)
                if (
                    k in candidate_conditions
                    and candidate_conditions[k]._values[1:] != new_condition._values[1:]  # type: ignore[attr-defined,union-attr]
                ):
                    raise DBQueryException(f"Multiple values provided for attribute {k}!")
                candidate_conditions[k] = Key(k).eq(v)
        elif len(_._values) and isinstance(_._values[0], (Key, Attr)):  # type: ignore[attr-defined,union-attr]
            attr_name = cast(str, _._values[0].name)  # type: ignore[attr-defined,union-attr]
            if attr_name not in index_all_key_names or not isinstance(
                _, SupportedKeyComparisonTypes
            ):
                non_candidate_conditions.append(_)
                continue
            if (
                attr_name in candidate_conditions
                and candidate_conditions[attr_name]._values[1:] != _._values[1:]  # type: ignore[union-attr]
            ):
                raise DBQueryException(f"Multiple values provided for attribute {attr_name}!")
            candidate_conditions[attr_name] = _
        else:
            non_candidate_conditions.append(_)

    for index in candidate_indexes:
        if index.key_name in candidate_conditions and isinstance(
            candidate_conditions[index.key_name], Equals
        ):
            target_index = index
            partition_key = candidate_conditions.pop(index.key_name)
            partition_key._values = (Key(index.key_name), *partition_key._values[1:])  # type: ignore[union-attr]
            if index.sort_key_name is not None and index.sort_key_name in candidate_conditions:
                sort_key_condition_expression = candidate_conditions.pop(index.sort_key_name)
                sort_key_condition_expression._values = (  # type: ignore[union-attr]
                    Key(index.sort_key_name),
                    *sort_key_condition_expression._values[1:],  # type: ignore[union-attr]
                )
            break

    # convert all remaining to filters
    filter_expressions = list(candidate_conditions.values()) + non_candidate_conditions

    return target_index, partition_key, sort_key_condition_expression, filter_expressions


@dataclass
class DynamoDBTable(LoggingMixin, Generic[DB_MODEL, DB_INDEX]):
    def __post_init__(self):
        check_table_name_and_index_match(self.table_name, self.get_db_index_cls())

    @property
    def table_name(self) -> str:
        return self.get_db_index_cls().table_name()

    def get_index_name(self, index: DB_INDEX | None = None) -> str | None:
        return self.index_or_default(index).index_name

    @classmethod
    @functools.cache
    def get_db_model_cls(cls) -> type[DB_MODEL]:
        return cls.__orig_bases__[0].__args__[0]  # type: ignore

    @classmethod
    @functools.cache
    def get_db_index_cls(cls) -> type[DB_INDEX]:
        return cls.__orig_bases__[0].__args__[1]  # type: ignore

    @classmethod
    def build_entry(cls, item: dict[str, Any], **kwargs) -> DB_MODEL:
        model_cls = cls.get_db_model_cls()
        return cast(DB_MODEL, model_cls.from_dict(item, **kwargs))

    @classmethod
    def build_item(cls, entry: DB_MODEL, partial: bool = False, **kwargs) -> dict[str, Any]:
        if partial:
            warnings.warn(
                "Deprecation Warning: partial=True is deprecated and will be removed in a future "
                "release. Please use the 'partial_model' method of your DBModel class to create "
                "a partial model class and use that as the DB_MODEL type parameter for your "
                "DynamoDBTable subclass.",
                DeprecationWarning,
                stacklevel=2,
            )
        # kwargs.setdefault("exclude_unset", partial)
        entry_dict = entry.to_dict(**kwargs)
        return convert_floats_to_decimals(entry_dict)

    @classmethod
    def index_or_default(cls, index: DB_INDEX | None = None) -> DB_INDEX:
        return index if index is not None else cls.get_db_index_cls().get_default_index()

    @classmethod
    def build_key_from_entry(cls, entry: DB_MODEL, index: DB_INDEX | None = None) -> DynamoDBKey:
        index = cls.index_or_default(index)
        if index.sort_key_name:
            return index.get_primary_key(
                partition_value=getattr(entry, index.key_name),
                sort_value=getattr(entry, index.sort_key_name),
            )
        else:
            return index.get_primary_key(partition_value=getattr(entry, index.key_name))

    @classmethod
    def build_key_from_item(
        cls, item: dict[str, Any], index: DB_INDEX | None = None
    ) -> DynamoDBKey:
        index = cls.index_or_default(index)
        if index.sort_key_name:
            return index.get_primary_key(
                partition_value=item.get(index.key_name), sort_value=item.get(index.sort_key_name)
            )
        else:
            return index.get_primary_key(partition_value=item.get(index.key_name))

    @classmethod
    def build_key(
        cls,
        key: DynamoDBItemValue | tuple[DynamoDBItemValue, DynamoDBItemValue] | DynamoDBKey,
        index: DB_INDEX | None = None,
    ) -> DynamoDBKey:
        if isinstance(key, MutableMapping):
            return key
        index = cls.index_or_default(index)
        return (
            index.get_primary_key(key[0], key[1])
            if isinstance(key, tuple)
            else index.get_primary_key(key)
        )

    # --------------------------------------------------------------------------
    # DB read methods (get, batch_get, query, scan, smart_query)
    # --------------------------------------------------------------------------
    def get(
        self,
        key: DynamoDBKey | DynamoDBItemValue | tuple[DynamoDBItemValue, DynamoDBItemValue],
        partial: bool = False,
    ) -> DB_MODEL:
        """Get a single item from a DynamoDB table by providing a key.

        Args:
            key: The partition key value that should be used to search the table.
                Can be a single partition key value, a tuple of partition and sort key value,
                or a dictionary of attribute:value.
            partial: Whether partial values are allowed. Defaults to False.

        Raises:
            DBReadException: If the item could not be found.

        Returns:
            The database entry that was found.
        """
        item_key = self.build_key(key)

        item = table_get_item(table_name=self.table_name, key=item_key)
        if item is None:
            raise DBReadException(f"Could not resolve item from {item_key}")
        return self.build_entry(item, partial=partial)

    def batch_get(
        self,
        keys: (
            Sequence[DynamoDBKey]
            | Sequence[DynamoDBItemValue]
            | Sequence[tuple[DynamoDBItemValue, DynamoDBItemValue]]
        ),
        ignore_missing: bool = False,
    ) -> list[DB_MODEL]:
        """Batch get items from a DynamoDB table by providing a list of keys.

        Args:
            keys: The partition key values that should be used to search the table.
                Each key can be one of:

                - **single partition key value**
                - **a tuple of partition and sort key value**
                - **a dictionary of attribute:value**

            ignore_missing: If true, suppress errors for keys that are not found.
                Defaults to False.

        Raises:
            DBReadException: If any of the items could not be found.

        Returns:
            List of database entries that were found.
        """
        if not keys:
            return []

        index = self.index_or_default()
        item_keys = [self.build_key(key, index=index) for key in keys]

        items = table_get_items(table_name=self.table_name, keys=item_keys)
        if len(items) != len(item_keys) and not ignore_missing:
            missing_keys = {
                (_[index.key_name], _.get(index.sort_key_name or "")) for _ in item_keys
            }.difference((_[index.key_name], _.get(index.sort_key_name or "")) for _ in items)

            raise DBReadException(f"Could not find items for {missing_keys}")
        entries = [self.build_entry(_) for _ in items]
        return entries

    def query(
        self,
        index: DB_INDEX,
        partition_key: DynamoDBPrimaryKeyItemValue | ConditionBase,
        sort_key_condition_expression: ConditionBase | None = None,
        filters: list[ConditionBase] | None = None,
        consistent_read: bool = False,
        expect_non_empty: bool = False,
        expect_unique: bool = False,
    ) -> list[DB_MODEL]:
        """Query a DynamoDB table by providing a DBIndex, partition_key, optional sort_key, and optional filter conditions.

        Args:
            index: Specifies the specific table index (e.g. main table, global secondary
                index, or local secondary index) that should be queried.
            partition_key: The partition key value that should be used to search the table.
            sort_key_condition_expression: The sort key condition expression to be used
                to query the table. Example:

                ```python
                from boto3.dynamodb.conditions import Key
                Key('my_sort_key_name').begins_with('prefix_')
                ```

            filters: A list of ConditionBase expressions where query results must satisfy.
            consistent_read: Whether a strongly consistent read should be used
                for the query. By default False which returns eventually consistent reads.
            expect_non_empty: Whether the resulting query should return at least
                one result. An error will be raised if expect_non_empty=True and 0 results were
                returned by the query.
            expect_unique: Whether the result of the query is expected to
                return AT MOST one result. An error will be raised if expect_unique=True and MORE
                than 1 result was returned for the query.

        Returns:
            A list of database model entries where partition_key/sort_key and
                filter conditions are satisfied.
        """  # noqa: E501

        if consistent_read:
            check_index_supports_strongly_consistent_read(index=index)

        index_name = self.get_index_name(index)

        key_condition_expression = self._build_key_condition_expression(
            index=index,
            partition_key=partition_key,
            sort_key_condition_expression=sort_key_condition_expression,
        )
        filter_expression = self._build_filter_condition_expression(filters=filters)

        self.log.info(
            f"Calling query on {self.table_name} table (index: {index_name}, "
            f"key condition: {condition_to_str(key_condition_expression)}, "
            f"filters: {condition_to_str(filter_expression)})"
        )

        items = table_query(
            table_name=self.table_name,
            index_name=index_name,
            key_condition_expression=key_condition_expression,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )
        if expect_unique:
            check_db_query_unique(
                index=index,
                query_result=items,
                key_condition_expression=key_condition_expression,
                filter_expression=filter_expression,
            )
        if expect_non_empty:
            check_db_query_non_empty(
                index=index,
                query_result=items,
                key_condition_expression=key_condition_expression,
                filter_expression=filter_expression,
            )
        entries = self._fill_values(items)
        return entries

    def scan(
        self,
        index: DB_INDEX | None = None,
        filters: list[ConditionBase] | None = None,
        consistent_read: bool = False,
        expect_non_empty: bool = False,
        expect_unique: bool = False,
    ) -> list[DB_MODEL]:
        """Scan a DynamoDB table by providing a DBIndex and optional filter conditions.

        Args:
            index: Specifies the specific table index (e.g. main table, global secondary
                index, or local secondary index) that should be scanned.
            filters: A list of ConditionBase expressions where scan results must satisfy.
            consistent_read: Whether a strongly consistent read should be used
                for the scan. By default False which returns eventually consistent reads.
            expect_non_empty: Whether the resulting scan should return at least
                one result. An error will be raised if expect_non_empty=True and 0 results were
                returned by the scan.
            expect_unique: Whether the result of the scan is expected to
                return AT MOST one result. An error will be raised if expect_unique=True and MORE
                than 1 result was returned for the scan.

        Returns:
            A list of database model entries where filter conditions are satisfied.
        """
        index = self.index_or_default(index)
        if consistent_read:
            check_index_supports_strongly_consistent_read(index=index)

        index_name = self.get_index_name(index)
        filter_expression = self._build_filter_condition_expression(filters=filters)

        self.log.info(
            f"Calling scan on {self.table_name} table (index: {index_name},"
            f" filters: {condition_to_str(filter_expression)})"
        )

        items = table_scan(
            table_name=self.table_name,
            index_name=index_name,
            filter_expression=filter_expression,
            consistent_read=consistent_read,
        )
        if expect_unique:
            check_db_query_unique(
                index=index,
                query_result=items,
                filter_expression=filter_expression,
            )
        if expect_non_empty:
            check_db_query_non_empty(
                index=index,
                query_result=items,
                filter_expression=filter_expression,
            )
        entries = self._fill_values(items)
        return entries

    def smart_query(
        self,
        *filters: DynamoDBKey | ConditionBase,
        consistent_read: bool = False,
        expect_non_empty: bool = False,
        expect_unique: bool = False,
        allow_scan: bool = True,
        **kw_filters: Any,
    ) -> list[DB_MODEL]:
        """
        Perform a smart query on the DynamoDB table by automatically determining whether to
        use a query or scan operation based on the provided filters.

        Args:
            *filters (Union[DynamoDBKey, ConditionBase]):
                Varargs of DynamoDBKey or ConditionBase to filter the query/scan
            consistent_read (bool):
                Whether a strongly consistent read should be used for the query/scan.
                Defaults to False which returns eventually consistent reads.
            expect_non_empty (bool):
                Whether the resulting query/scan should return at least one result.
                An error will be raised if expect_non_empty=True and 0 results were returned.
            expect_unique (bool):
                Whether the result of the query/scan is expected to return AT MOST one result.
                An error will be raised if expect_unique=True and MORE than 1 result was returned.
            allow_scan (bool):
                Whether to allow a scan operation if no partition key is found in the filters.
                Defaults to True.

        Raises:
            DBQueryException:
                If no partition key is found and allow_scan is False.

        Returns:
            A list of database model entries where partition_key/sort_key and filter
                conditions are satisfied.
        """  # noqa: E501
        (
            index,
            partition_key,
            sort_key_condition_expression,
            filter_expressions,
        ) = build_optimized_condition_expression_set(
            self.get_db_index_cls(),
            *filters,
            **kw_filters,
        )
        if partition_key:
            return self.query(
                index=self.index_or_default(index),
                partition_key=partition_key,
                sort_key_condition_expression=sort_key_condition_expression,
                filters=filter_expressions,
                consistent_read=consistent_read,
                expect_non_empty=expect_non_empty,
                expect_unique=expect_unique,
            )
        else:
            if not allow_scan:
                raise DBQueryException(
                    f"Could not find a partition key for {self.table_name} table! "
                    "Please provide a partition key or allow_scan=True."
                )
            if sort_key_condition_expression:
                filter_expressions.append(sort_key_condition_expression)
            return self.scan(
                index=index,
                filters=filter_expressions,
                consistent_read=consistent_read,
                expect_non_empty=expect_non_empty,
                expect_unique=expect_unique,
            )

    # --------------------------------------------------------------------------
    # DB write methods (put, update, delete)
    # --------------------------------------------------------------------------
    def put(
        self,
        entry: DB_MODEL,
        condition_expression: ConditionBase | None = None,
        **table_put_item_kwargs,
    ) -> DB_MODEL:
        """Put a new item into the DynamoDB table.

        Args:
            entry (DB_MODEL): The database model entry to be put into the table.
            condition_expression (Optional[ConditionBase], optional): An optional condition
                expression that must be satisfied for the put operation to succeed.
                Defaults to None.
            **table_put_item_kwargs: Additional keyword arguments to be passed to the
                table_put_item function.

        Raises:
            DBWriteException: If there was an error putting the entry.
            DBWriteException: If the HTTP response code indicates a failure.

        Returns:
            The database model entry that was put into the table.
        """
        put_summary = (
            f"(entry: {entry}, condition_expression: {condition_to_str(condition_expression)})"
        )
        self.log.debug(f"{self.table_name} - Putting new entry: {put_summary}")

        e_msg_intro = f"{self.table_name} - Error putting entry: {put_summary}."
        try:
            put_response = table_put_item(
                table_name=self.table_name,
                item=self.build_item(entry),
                condition_expression=condition_expression,
                **table_put_item_kwargs,
            )
        except ClientError as e:
            e_msg_client = f"{e_msg_intro} Details: {get_client_error_code(e)}"
            self.log.error(e_msg_client)
            raise DBWriteException(e_msg_client)
        else:
            if not (200 <= put_response["ResponseMetadata"]["HTTPStatusCode"] < 300):
                e_msg_http = f"{e_msg_intro} Table put_item response: {put_response}"
                raise DBWriteException(e_msg_http)
        self.log.debug(f"{self.table_name} - Put successful: {put_response}")
        return entry

    def update(
        self,
        key: DynamoDBKey,
        new_entry: Mapping[str, Any] | DB_MODEL,
        old_entry: DB_MODEL | None = None,
        **table_update_item_kwargs,
    ) -> DB_MODEL:
        """Update an existing item in the DynamoDB table.

        Args:
            key (DynamoDBKey): The primary key of the item to be updated.
            new_entry (Union[Mapping[str, Any], DB_MODEL]): The new values for the item.
                Can be a dictionary of attribute:value pairs or a DB_MODEL instance.
            old_entry (Optional[DB_MODEL], optional): The existing entry before the update.
                If provided, only attributes that differ from the old_entry will be updated.
                Defaults to None.
            **table_update_item_kwargs: Additional keyword arguments to be passed to the
                table_update_item function.

        Raises:
            DBWriteException: If there was an error updating the entry.
            DBWriteException: If no attributes need to be updated.

        Returns:
            The updated database model entry.
        """
        new_attributes: dict[str, Any] = {}
        if isinstance(new_entry, self.get_db_model_cls()):
            new_attributes = self.build_item(new_entry, partial=True)
        elif new_entry:
            # we still need to convert floats to decimals if new_entry is a dict
            new_attributes = convert_floats_to_decimals(new_entry)  # type: ignore[arg-type]

        for k in key:
            new_attributes.pop(k, None)
        # Add k:v pair from new_attributes if new != old value for a given key
        new_clean_attrs: dict[str, Any] = {}
        if old_entry:
            for k, new_v in new_attributes.items():
                if getattr(old_entry, k) != new_v:
                    new_clean_attrs[k] = new_v
        else:
            new_clean_attrs = new_attributes

        if not new_clean_attrs:
            self.log.debug(
                f"{self.table_name} - No attr_updates to do! Skipping _update_entry call."
            )
            if not old_entry:
                old_entry = self.get(key)
            return old_entry

        update_summary = f"(old_entry: {old_entry}, new_attributes: {new_clean_attrs})"
        self.log.debug(f"{self.table_name} - Updating entry: {update_summary}")
        try:
            updated_item = table_update_item(
                table_name=self.table_name,
                key=key,
                attributes=new_clean_attrs,
                return_values="ALL_NEW",
                **table_update_item_kwargs,
            )
            # table_update_item will always return a dict if ReturnValues != "NONE"
            if updated_item is None:
                raise DBWriteException(
                    f"{self.table_name} - Error updating entry: {update_summary}"
                )
            updated_entry = self.build_entry(updated_item)
        except ClientError as e:
            e_msg = (
                f"{self.table_name} - Error updating entry: {update_summary}. "
                f"Details: {get_client_error_code(e)}"
            )
            self.log.error(e_msg)
            raise DBWriteException(e_msg)
        self.log.debug(f"{self.table_name} - Successfully updated entry: {updated_entry}")
        return updated_entry

    @overload
    def delete(
        self,
        key: DynamoDBKey | DB_MODEL,
        error_on_nonexistent: Literal[True],
    ) -> DB_MODEL: ...

    @overload
    def delete(
        self,
        key: DynamoDBKey | DB_MODEL,
        error_on_nonexistent: Literal[False],
    ) -> DB_MODEL | None: ...

    @overload
    def delete(
        self,
        key: DynamoDBKey | DB_MODEL,
    ) -> DB_MODEL | None: ...

    def delete(
        self,
        key: DynamoDBKey | DB_MODEL,
        error_on_nonexistent: bool = False,
    ) -> DB_MODEL | None:
        """Delete an item from the DynamoDB table.

        Args:
            key (Union[DynamoDBKey, DB_MODEL]): The primary key of the item to be deleted,
                or a DB_MODEL instance representing the item to be deleted.
            error_on_nonexistent (bool, optional): Whether to raise an error if the item
                does not exist. Defaults to False.

        Raises:
            DBWriteException: If there was an error deleting the entry.
            DBWriteException: If error_on_nonexistent is True and the item does not exist.

        Returns:
            The deleted database model entry, or None if the item did not exist and
                error_on_nonexistent is False.
        """
        if isinstance(key, self.get_db_model_cls()):
            key = self.build_key_from_entry(key)
        delete_summary = f"(db_primary_key: {key})"
        self.log.debug(f"{self.table_name} - Deleting entry with: {delete_summary}")
        e_msg = f"{self.table_name} - Delete failed for the following primary key: {key}"
        try:
            deleted_attributes = table_delete_item(
                table_name=self.table_name,
                key=cast(DynamoDBKey, key),
                return_values="ALL_OLD",  # type: ignore[arg-type] # expected type more general than specified here
            )

            if not deleted_attributes:
                self.log.info(f"{self.table_name} - Nothing deleted for primary key: {key}")
                if error_on_nonexistent:
                    raise DBWriteException(e_msg)
                return None
            else:
                deleted_entry = self.build_entry(deleted_attributes)
                self.log.info(f"{self.table_name} - Deleted entry: {deleted_attributes}")
                return deleted_entry
        except ClientError as e:
            detailed_e_msg = f"{e_msg}. Details: {get_client_error_code(e)}"
            raise DBWriteException(detailed_e_msg)

    # --------------------------------------------------------------------------
    # Key and filter utils
    # --------------------------------------------------------------------------
    def _build_key_condition_expression(
        self,
        index: DB_INDEX,
        partition_key: DynamoDBPrimaryKeyItemValue | ConditionBase,
        sort_key_condition_expression: ConditionBase | None = None,
    ) -> ConditionBase:
        partition_key_name = index.key_name
        sort_key_name = index.sort_key_name

        # Build dynamodb key condition expression
        assert partition_key_name is not None
        key_condition_expression: ConditionBase
        if isinstance(partition_key, ConditionBase):
            key_condition_expression = partition_key
        else:
            key_condition_expression = Key(partition_key_name).eq(partition_key)

        if sort_key_condition_expression is not None:
            condition_key_name = sort_key_condition_expression.get_expression()["values"][0].name
            if sort_key_name is not None:
                if sort_key_name == condition_key_name:
                    key_condition_expression &= sort_key_condition_expression
                else:
                    raise DBQueryException(
                        "The sort key specified by the provided sort_key_condition_expression "
                        f"({condition_key_name}) does not match the sort key name of the index "
                        f"to be queried (index: {index}, index_sort_key_name: {sort_key_name})!"
                    )
            else:
                self.log.warning(
                    f"A sort key condition expression was provided "
                    f"({sort_key_condition_expression}) for the query but the specified "
                    f"table index {index} does not support a sort key!"
                )
        return key_condition_expression

    def _build_filter_condition_expression(
        self, filters: list[ConditionBase] | None
    ) -> ConditionBase | None:
        # Build dynamodb attribute condition expression
        filter_expression: ConditionBase | None = None
        if filters:
            filter_expression = functools.reduce(lambda a, b: a & b, filters)
        return filter_expression

    def _fill_values(self, items: Sequence[dict[str, Any]]) -> list[DB_MODEL]:
        HashableKey: TypeAlias = tuple[tuple[str, DynamoDBPrimaryKeyItemValue], ...]

        db_model_cls = self.get_db_model_cls()
        filled_entries: list[DB_MODEL | None] = [None] * len(items)
        partial_keys: list[DynamoDBKey] = []
        key_to_index: dict[HashableKey, int] = {}

        for i, item in enumerate(items):
            try:
                filled_entries[i] = db_model_cls.from_dict(item)
            except ValidationError:
                key = self.build_key_from_item(item)
                partial_keys.append(key)
                key_to_index[tuple(sorted(key.items()))] = i  # type: ignore[arg-type]

        if partial_keys:
            for refetched in self.batch_get(partial_keys):
                refetched_key = self.build_key_from_item(refetched.to_dict())
                hk = tuple(sorted(refetched_key.items()))
                if hk in key_to_index:
                    filled_entries[key_to_index[hk]] = refetched  # type: ignore[index]

        return cast(list[DB_MODEL], filled_entries)

    def execute_partiql_statement(self, statement: str) -> Sequence:
        return execute_partiql_statement(statement=statement)

    @classmethod
    def from_env(
        cls: type["DynamoDBTable[DB_MODEL, DB_INDEX]"], *args, **kwargs
    ) -> "DynamoDBTable[DB_MODEL, DB_INDEX]":
        return cls(*args, **kwargs)


@dataclass
class DynamoDBEnvBaseTable(DynamoDBTable[DB_MODEL, DB_INDEX], Generic[DB_MODEL, DB_INDEX]):
    env_base: EnvBase = field(default_factory=EnvBase.from_env)

    @property
    def table_name(self) -> str:
        return self.env_base.get_table_name(super().table_name)

    def get_index_name(self, index: DB_INDEX | None = None) -> str | None:
        if (index_name := super().get_index_name(index)) is not None:
            return self.env_base.prefixed(index_name)
        return index_name



from __future__ import annotations

import re
from functools import cache

__all__ = [
    "MountPointConfiguration",
    "detect_mount_points",
    "deduplicate_mount_points",
]

import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Union, cast

from aibs_informatics_core.models.aws.efs import AccessPointId, EFSPath, FileSystemId
from aibs_informatics_core.utils.hashing import sha256_hexdigest
from aibs_informatics_core.utils.os_operations import get_env_var
from botocore.exceptions import NoCredentialsError
from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_exponential_jitter

from aibs_informatics_aws_utils.constants.efs import (
    EFS_MOUNT_POINT_ID_VAR,
    EFS_MOUNT_POINT_PATH_VAR,
)
from aibs_informatics_aws_utils.core import AWSService
from aibs_informatics_aws_utils.efs.core import get_efs_access_point, get_efs_file_system

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_batch.type_defs import JobDetailTypeDef
    from mypy_boto3_efs.type_defs import (
        AccessPointDescriptionTypeDef,
        FileSystemDescriptionTypeDef,
    )
else:
    AccessPointDescriptionTypeDef = dict
    FileSystemDescriptionTypeDef = dict
    JobDetailTypeDef = dict

logger = logging.getLogger(__name__)

StrPath = Union[Path, str]


get_lambda_client = AWSService.LAMBDA.get_client
get_batch_client = AWSService.BATCH.get_client


@dataclass
class MountPointConfiguration:
    """
    Adapter for translating and mapping paths between AWS Elastic File System
    (EFS) mount points and local file system paths.

    local file systems are most often used by:
    - EC2 instances
    - AWS Lambda functions

    This class provides functionality to adapt file paths for applications
    that interact with AWS EFS. It allows for the conversion of paths from the
    EFS perspective to the local file system perspective and vice versa,
    considering the mount point and access point configurations.

    Attributes:
        file_system (FileSystemDescriptionTypeDef): The description of the EFS file system.
        access_point (Optional[AccessPointDescriptionTypeDef]): The description
            of the EFS access point, if used.
        mount_point (Path): The local file system path where the EFS is mounted.
    """

    file_system: FileSystemDescriptionTypeDef
    access_point: AccessPointDescriptionTypeDef | None
    mount_point: Path

    # TODO: make hashable

    @property
    def access_point_path(self) -> Path:
        """This is the access point on file system mounted to host"""
        path: str = "/"
        if self.access_point:
            path = self.access_point.get("RootDirectory", {}).get("Path", path)
        return Path(path)

    @property
    def mount_point_path(self) -> Path:
        """This is the path on the host where the access point will be mounted"""
        return self.mount_point

    @property
    def is_writeable(self) -> bool:
        """Returns true, if the mount point is writeable"""
        # TODO: need to figure out if it is writeable
        return True

    def translate_mounted_path(self, path: StrPath, other: MountPointConfiguration) -> Path:
        """
        Translates the location of a path described by another mount point to
        a location on this mount point

        Args:
            path (StrPath): The path to translate. should be relative path
                or absolute relative to the other mount point/access point.
            other (T): The config of other mount point/access point to translate from.

        Raises:
            ValueError: If file system of other mount point/access point is
                not the same as this mount point/access point.

        Returns:
            The translated absolute path on this mount point.
        """
        path = Path(path)
        if self.file_system["FileSystemId"] != other.file_system["FileSystemId"]:
            raise ValueError(
                f"Cannot resolve path {path} from file system "
                f"{other.file_system['FileSystemId']} to file system "
                f"{self.file_system['FileSystemId']}. They are not the same file system!"
            )

        # First, convert the path to a global path on EFS File System
        global_path = other.as_efs_path(path)
        # Next, convert the global path to the path on the container
        return self.as_mounted_path(global_path)

    def as_relative_path(self, path: StrPath) -> Path:
        """
        Converts a path to a relative path from the mount point or access point.

        Behavior:
        - If the path is absolute and is relative to the mount point or access point,
            it is returned as a relative path.
        - If the path is already relative,
            it is returned as is.
        - If the path is absolute and is not relative to the mount point or access point,
            a ValueError is raised.


        Args:
            path (StrPath): The path to convert to a relative path.

        Returns:
            The relative path.

        """
        path = Path(path)
        if path.is_absolute():
            # Check longer path first. This is to avoid the case where the mount point
            # is a subdirectory of the access point.
            check_mount_path_first = len(self.mount_point_path.as_posix()) > len(
                self.access_point_path.as_posix()
            )
            if check_mount_path_first and self.is_mounted_path(path):
                path = path.relative_to(self.mount_point)
            elif self.is_efs_path(path):
                path = path.relative_to(self.access_point_path)
            elif self.is_mounted_path(path):
                path = path.relative_to(self.mount_point)
            else:
                raise ValueError(
                    f"Path {path} is not relative to either the mount point {self.mount_point} "
                    f"or the access point {self.access_point_path}"
                )
        return path

    def as_efs_path(self, path: StrPath) -> Path:
        """
        Converts a path to a path on the EFS file system.

        Behavior:
        - If the path is absolute, the path is made relative to either the
            mount point or access point first. examples:
            - "/efs/accesspoint/path/to/file" -> "path/to/file"
            - "/mnt/efs/path/to/file" -> "path/to/file"
        - If the path is absolute and does not start with the mount point or access point,
            a ValueError is raised.
        - If the path is relative, It is assumed to be the relative path from the access point.
            e.g. "path/to/file" -> "/efs/accesspoint/path/to/file"

        Args:
            path (StrPath): The path to convert to a path on the EFS file system.

        Returns:
            absolute path on the EFS file system.
        """
        path = self.as_relative_path(path)
        return (self.access_point_path / path).resolve()

    def as_efs_uri(self, path: StrPath) -> EFSPath:
        """
        Converts a path to a customized URI EFSPath describing the location on an EFS file system.

        Args:
            path (StrPath): The path to convert to a path on the EFS file system.

        Returns:
            absolute URI path on the EFS file system.
        """
        return EFSPath.build(
            resource_id=self.file_system["FileSystemId"],
            path=self.as_efs_path(path),
        )

    def as_mounted_path(self, path: StrPath) -> Path:
        """
        Converts a path to a path on the host where the EFS is mounted.

        Behavior:
        - If the path is absolute, the path is made relative to either the mount point
            or access point first. examples:
            - "/efs/accesspoint/path/to/file" -> "path/to/file"
            - "/mnt/efs/path/to/file" -> "path/to/file"
        - If the path is absolute and does not start with the mount point or access point,
            a ValueError is raised.
        - If the path is relative, It is assumed to be the relative path from the mount point.
            e.g. "path/to/file" -> "/mnt/efs/path/to/file"

        Args:
            path (StrPath): The path to convert to a path on the host where the EFS is mounted.

        Returns:
            absolute path on the host where the EFS is mounted.
        """
        path = self.as_relative_path(path)
        return (self.mount_point / path).resolve()

    def is_efs_path(self, path: StrPath) -> bool:
        """Checks if the path is relative to the access point"""
        return Path(path).is_relative_to(self.access_point_path)

    def is_mounted_path(self, path: StrPath) -> bool:
        """Checks if the path is relative to the mount point"""
        return Path(path).is_relative_to(self.mount_point)

    def as_env_vars(self, name: str | None = None) -> dict[str, str]:
        """Converts the mount point configuration to environment variables."""
        if self.access_point and self.access_point.get("AccessPointId"):
            mount_point_id = self.access_point["AccessPointId"]
        else:
            mount_point_id = self.file_system["FileSystemId"]

        return self.to_env_vars(self.mount_point, mount_point_id, name)

    @classmethod
    def build(
        cls,
        mount_point: StrPath,
        access_point: str | AccessPointDescriptionTypeDef | None = None,
        file_system: str | FileSystemDescriptionTypeDef | None = None,
        access_point_tags: dict[str, str] | None = None,
        file_system_tags: dict[str, str] | None = None,
    ) -> MountPointConfiguration:
        """Creates a new config from the given mount point and access point or file system.

        Important: Must provide either access point or file system.

        Args:
            mount_point (StrPath): Intended mount point of the EFS file system on the host.
            access_point (Optional[Union[str, AccessPointDescriptionTypeDef]]):
                Identifier of the access point or the access point description.
                If specified as a string, can be either the access point id or name.
                Defaults to None.
            file_system (Optional[Union[str, FileSystemDescriptionTypeDef]]):
                Identifier of the file system or the file system description.
                If specified as a string, can be either the file system id or name.
                Defaults to None.
            access_point_tags (Optional[Dict[str, str]]): Tags to filter the access point.
                Defaults to None.
            file_system_tags (Optional[Dict[str, str]]): Tags to filter the file system.
                Defaults to None.
        Raises:
            ValueError: if neither access point nor file system is provided.

        Returns:
            The config
        """

        if access_point is None and file_system is None:
            raise ValueError("Either access point or file system must be provided")

        file_system_id = None
        if isinstance(file_system, str):
            logger.info(f"Resolving file system config from name {file_system}")
            fs_id, fs_name = (
                (FileSystemId(file_system).normalized, None)
                if FileSystemId.is_valid(file_system)
                else (None, file_system)
            )
            file_system = get_efs_file_system(
                name=fs_name, file_system_id=fs_id, tags=file_system_tags
            )
            file_system_id = file_system["FileSystemId"]
            logger.info(f"Resolved file system id {file_system_id}")

        access_point_id = None
        if isinstance(access_point, str):
            logger.info(f"Resolving access point config from name {access_point}")
            ap_id, ap_name = (
                (AccessPointId(access_point).normalized, None)
                if AccessPointId.is_valid(access_point)
                else (None, access_point)
            )
            access_point = get_efs_access_point(
                access_point_name=ap_name,
                access_point_id=ap_id,
                file_system_id=file_system_id,
                access_point_tags=access_point_tags,
            )
            access_point_id = access_point.get("AccessPointId")
            logger.info(f"Resolved access point id {access_point_id}")

        if file_system is None:
            assert access_point is not None
            file_system_id = access_point.get("FileSystemId")
            logger.info(f"Resolving file system config from access point id {access_point_id}")
            file_system = get_efs_file_system(file_system_id=access_point.get("FileSystemId"))
            logger.info(f"Resolved file system id {file_system_id}")

        if not isinstance(mount_point, Path):
            mount_point = Path(mount_point)

        logger.info(
            f"Generating EFS Mount Point Connection using "
            f"file system {file_system}, "
            f"access point {access_point}, "
            f"with target {mount_point} mount point"
        )
        return cls(
            file_system=file_system,
            access_point=access_point,
            mount_point=mount_point,
        )

    @classmethod
    def to_env_vars(
        cls,
        mount_point: StrPath,
        mount_point_id: str | AccessPointId | FileSystemId,
        name: str | None = None,
    ) -> dict[str, str]:
        """Converts the mount point configuration to environment variables.

        Args:
            mount_point (StrPath): The mount point.
            mount_point_id (Union[str, AccessPointId, FileSystemId]): The mount point identifier.
            name (Optional[str]): Optional name for the mount point. Defaults to None.

        Returns:
            Dict[str, str]: The environment variables.
        """

        if not isinstance(mount_point, Path):
            mount_point = Path(mount_point)

        if not isinstance(mount_point_id, AccessPointId) and not isinstance(
            mount_point_id, FileSystemId
        ):
            if AccessPointId.is_valid(mount_point_id):
                mount_point_id = AccessPointId(mount_point_id).normalized
            elif FileSystemId.is_valid(mount_point_id):
                mount_point_id = FileSystemId(mount_point_id).normalized
            else:
                raise ValueError(
                    f"Invalid mount point id {mount_point_id}. "
                    "Must be either an access point id or file system id."
                )
        else:
            mount_point_id = mount_point_id.normalized

        if name and not re.match(r"[a-zA-Z][\w]{0,11}", name):
            raise ValueError(
                f"Invalid mount point name {name}. Must be a valid environment variable name."
            )
        else:
            name = name or sha256_hexdigest([mount_point.as_posix(), mount_point_id])[:6]

        return {
            f"{EFS_MOUNT_POINT_PATH_VAR}_{name}": mount_point.as_posix(),
            f"{EFS_MOUNT_POINT_ID_VAR}_{name}": mount_point_id,
        }

    def __repr__(self) -> str:
        access_point = self.access_point.get("AccessPointId") if self.access_point else None
        return (
            f"{self.__class__.__name__}(file_system={self.file_system['FileSystemId']}, "
            f"access_point={access_point},"
            f"access_point_path={self.access_point_path}, mount_point={self.mount_point})"
        )


@cache
@retry(
    retry=retry_if_exception_type(NoCredentialsError),
    stop=stop_after_attempt(5),
    wait=wait_exponential_jitter(initial=1, exp_base=2, jitter=1),
    reraise=True,
)
def detect_mount_points() -> list[MountPointConfiguration]:
    mount_points: list[MountPointConfiguration] = []

    if batch_job_id := get_env_var("AWS_BATCH_JOB_ID"):
        logger.info(f"Detected Batch job {batch_job_id}")
        batch_mp_configs = _detect_mount_points_from_batch_job(batch_job_id)
        logger.info(f"Detected {len(batch_mp_configs)} EFS mount points from Batch")
        mount_points.extend(batch_mp_configs)
    elif lambda_function_name := get_env_var("AWS_LAMBDA_FUNCTION_NAME"):
        logger.info(f"Detected Lambda function {lambda_function_name}")
        lambda_mp_configs = _detect_mount_points_from_lambda(lambda_function_name)
        logger.info(f"Detected {len(lambda_mp_configs)} EFS mount points from Lambda")
        mount_points.extend(lambda_mp_configs)
    else:
        logger.info("No Lambda or Batch environment detected. Using environment variables.")
        env_mount_points = _detect_mount_points_from_env()
        logger.info(
            f"Detected {len(env_mount_points)} EFS mount points from environment variables"
        )
        mount_points.extend(env_mount_points)

    logger.info(f"Detected {len(mount_points)} EFS mount points. Deuplicating...")
    mount_points = deduplicate_mount_points(mount_points)
    return mount_points


def deduplicate_mount_points(
    mount_points: list[MountPointConfiguration],
) -> list[MountPointConfiguration]:
    """
    Deduplicates a list of MountPointConfiguration objects based on the file
    system id and access point id.
    """

    unique_configs: dict[str, MountPointConfiguration] = {}
    for mp_config in mount_points:
        key = mp_config.mount_point.as_posix()

        if key in unique_configs:
            other = unique_configs[key]
            if mp_config.access_point_path != other.access_point_path:
                raise ValueError(
                    f"Found conflicting mount points for {key}: "
                    f"{mp_config.access_point_path} and {unique_configs[key].access_point_path}"
                )
            elif mp_config.file_system["FileSystemId"] != other.file_system["FileSystemId"]:
                raise ValueError(
                    f"Found conflicting file systems for {key}: "
                    f"{mp_config.file_system['FileSystemId']} "
                    f"and {unique_configs[key].file_system['FileSystemId']}"
                )
            elif mp_config.access_point != other.access_point:
                raise ValueError(
                    f"Found conflicting access points for {key}: "
                    f"{mp_config.access_point} and {unique_configs[key].access_point}"
                )
            continue
        unique_configs[key] = mp_config
    return list(unique_configs.values())


# ------------------------------------
# Private Helpers
# ------------------------------------


def _detect_mount_points_from_lambda(lambda_function_name: str) -> list[MountPointConfiguration]:
    mount_points: list[MountPointConfiguration] = []
    lambda_ = get_lambda_client()
    response = lambda_.get_function_configuration(FunctionName=lambda_function_name)

    fs_configs = response.get("FileSystemConfigs")
    if fs_configs:
        for fs_config in fs_configs:
            mount_points.append(
                MountPointConfiguration.build(
                    mount_point=fs_config["LocalMountPath"],
                    access_point=fs_config["Arn"],
                )
            )
    return _remove_invalid_mount_points(mount_points)


def _detect_mount_points_from_batch_job(batch_job_id: str) -> list[MountPointConfiguration]:
    mount_points: list[MountPointConfiguration] = []
    batch = get_batch_client()
    response = batch.describe_jobs(jobs=[batch_job_id])
    job_container = response.get("jobs", [cast(JobDetailTypeDef, {})])[0].get("container", {})
    batch_mount_points = job_container.get("mountPoints")
    batch_volumes = job_container.get("volumes")
    if batch_mount_points and batch_volumes:
        volume_mapping = {
            v_config["name"]: v_config
            for v_config in batch_volumes
            if "efsVolumeConfiguration" in v_config and "name" in v_config
        }

        for batch_mount_point in batch_mount_points:
            assert "containerPath" in batch_mount_point, "containerPath is required"
            assert "sourceVolume" in batch_mount_point, "sourceVolume is required"

            if (batch_volume := volume_mapping.get(batch_mount_point["sourceVolume"])) is None or (
                efs_vol_config := batch_volume.get("efsVolumeConfiguration")
            ) is None:
                continue

            mount_path = batch_mount_point["containerPath"]
            access_point = efs_vol_config.get("authorizationConfig", {}).get("accessPointId")
            file_system = efs_vol_config["fileSystemId"]

            mount_points.append(
                MountPointConfiguration.build(
                    mount_point=mount_path,
                    access_point=access_point,
                    file_system=file_system,
                )
            )

    return _remove_invalid_mount_points(mount_points)


def _detect_mount_points_from_env() -> list[MountPointConfiguration]:
    mount_points: list[MountPointConfiguration] = []

    for k, v in os.environ.items():
        if k.startswith(EFS_MOUNT_POINT_PATH_VAR):
            try:
                name = k[len(EFS_MOUNT_POINT_PATH_VAR) :]
                mount_point = Path(v)

                # if not mount_point.absolute():
                #     logger.warning(f"Mount point {v} is not an absolute path. Skipping...")
                #     continue
                # if not mount_point.is_dir():
                #     logger.warning(f"Mount point {v} is not a valid directory. Skipping...")
                #     continue

                if (efs_id := get_env_var(f"{EFS_MOUNT_POINT_ID_VAR}{name}")) is None:
                    logger.warning(f"Couldn't resolve EFS ID associated with {k} (at {v})")
                    continue

                file_system_id = FileSystemId.find_suffix(efs_id)
                access_point_id = AccessPointId.find_suffix(efs_id)

                if file_system_id is None and access_point_id is None:
                    logger.warning(f"Could not resolve file system or access point from {efs_id}")
                    continue

                mount_points.append(
                    MountPointConfiguration.build(
                        mount_point=mount_point,
                        access_point=access_point_id,
                        file_system=file_system_id,
                    )
                )
            except Exception as e:
                logger.warning(f"Error building config for EFS mount point {k}={v}: {e}")
    return _remove_invalid_mount_points(mount_points)


def _remove_invalid_mount_points(
    mount_points: list[MountPointConfiguration],
) -> list[MountPointConfiguration]:
    valid_mount_points: list[MountPointConfiguration] = []
    for mp in mount_points:
        try:
            if not mp.mount_point.absolute():
                raise ValueError(
                    f"Mount point {mp.mount_point} is not an absolute path. Skipping..."
                )
            if not mp.mount_point.is_dir():
                raise ValueError(
                    f"Mount point {mp.mount_point} is not a valid directory. Skipping..."
                )
            valid_mount_points.append(mp)
        except Exception as e:
            logger.warning(f"Error validating EFS mount point {mp.mount_point}: {e}")
    return valid_mount_points



from dataclasses import dataclass


@dataclass
class EFSTag:
    key: str
    value: str


## EFS Environment Variable Constants
EFS_MOUNT_POINT_PATH_VAR = "EFS_MOUNT_POINT_PATH"
EFS_MOUNT_POINT_ID_VAR = "EFS_MOUNT_POINT_ID"


EFS_MOUNT_POINT_PATH_VAR_PREFIX = "EFS_MOUNT_POINT_PATH_"
EFS_MOUNT_POINT_ID_VAR_PREFIX = "EFS_MOUNT_POINT_ID_"


# ------------------------------------
# Standard Names and Paths for EFS


# fmt: off
EFS_ROOT_PATH           = "/"
EFS_SHARED_PATH         = "/shared"
EFS_SCRATCH_PATH        = "/scratch"
EFS_TMP_PATH            = "/tmp"
# fmt: on


# fmt: off
EFS_ROOT_ACCESS_POINT_NAME      = "root"
EFS_SHARED_ACCESS_POINT_NAME    = "shared"
EFS_SCRATCH_ACCESS_POINT_NAME   = "scratch"
EFS_TMP_ACCESS_POINT_NAME       = "tmp"
# fmt: on


# fmt: off
EFS_ROOT_ACCESS_POINT_TAG       = EFSTag("Name", EFS_ROOT_ACCESS_POINT_NAME)
EFS_SHARED_ACCESS_POINT_TAG     = EFSTag("Name", EFS_SHARED_ACCESS_POINT_NAME)
EFS_SCRATCH_ACCESS_POINT_TAG    = EFSTag("Name", EFS_SCRATCH_ACCESS_POINT_NAME)
EFS_TMP_ACCESS_POINT_TAG        = EFSTag("Name", EFS_TMP_ACCESS_POINT_NAME)
# fmt: on



import base64
import json
import logging
import re
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypedDict,
    TypeVar,
    cast,
)

import requests
from aibs_informatics_core.collections import StrEnum, ValidatedStr
from aibs_informatics_core.models.base import PydanticBaseModel
from botocore.exceptions import ClientError
from pydantic import (
    Field,
    PrivateAttr,
    model_serializer,
    model_validator,
)

from aibs_informatics_aws_utils.core import (
    AWSService,
    get_account_id,
    get_client_error_code,
    get_region,
)
from aibs_informatics_aws_utils.exceptions import (
    AWSError,
    InvalidAmazonResourceNameError,
    ResourceNotFoundError,
)

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ecr import ECRClient
    from mypy_boto3_ecr.literals import ImageTagMutabilityType, TagStatusType
    from mypy_boto3_ecr.type_defs import (
        BatchGetImageResponseTypeDef,
        ImageDetailTypeDef,
        ImageIdentifierTypeDef,
        LayerTypeDef,
        ListImagesFilterTypeDef,
        ListImagesRequestPaginateTypeDef,
        TagTypeDef,
    )
else:
    ECRClient = object
    ImageTagMutabilityType = object
    TagStatusType = object
    BatchGetImageResponseTypeDef = dict
    ImageDetailTypeDef = dict
    ImageIdentifierTypeDef = dict
    LayerTypeDef = dict
    ListImagesFilterTypeDef = dict
    ListImagesRequestPaginateTypeDef = dict
    TagTypeDef = dict

logger = logging.getLogger(__name__)


get_ecr_client = AWSService.ECR.get_client

ECR_REPO_ARN_PATTERN = re.compile(
    r"arn:aws:ecr:([\w-]*):([\d]{10,12}):repository/((?:[a-z0-9]+(?:[._-][a-z0-9]+)*/)*[a-z0-9]+(?:[._-][a-z0-9]+)*)"
)

ECR_REGISTRY_PATTERN_STR = r"([\d]{12}).dkr.ecr.([\w-]*).amazonaws.com"
ECR_REPO_PATTERN_STR = r"/((?:[a-z0-9]+(?:[._-][a-z0-9]+)*/)*[a-z0-9]+(?:[._-][a-z0-9]+)*)"
ECR_IMAGE_PATTERN_STR = r"(:([\w.\-_]{1,127})|@(sha256:[A-Fa-f0-9]{64}))"

ECR_REGISTRY_URI_PATTERN = re.compile(ECR_REGISTRY_PATTERN_STR)
ECR_REPO_URI_PATTERN = re.compile(ECR_REGISTRY_PATTERN_STR + ECR_REPO_PATTERN_STR)
ECR_IMAGE_URI_PATTERN = re.compile(
    ECR_REGISTRY_PATTERN_STR + ECR_REPO_PATTERN_STR + ECR_IMAGE_PATTERN_STR
)


class ECRRegistryUri(ValidatedStr):
    regex_pattern = ECR_REGISTRY_URI_PATTERN

    @property
    def account_id(self) -> str:
        return self.get_match_groups()[0]

    @property
    def region(self) -> str:
        return self.get_match_groups()[1]

    @classmethod
    def from_components(
        cls, account_id: str | None = None, region: str | None = None, **kwargs
    ) -> "ECRRegistryUri":
        """Generate a Registry URI.

        If account ID not provided, account Id of credentials is used.
        If region is not provided, region of credentials is used.

        Args:
            account_id: The registry ID. Defaults to None.
            region: AWS region. Defaults to None.
            **kwargs: Additional keyword arguments (unused).

        Returns:
            AWS Registry URI.
        """
        account_id = account_id or get_account_id()
        region = get_region(region)
        return ECRRegistryUri(f"{account_id}.dkr.ecr.{region}.amazonaws.com")


class ECRRepositoryUri(ECRRegistryUri):
    regex_pattern = ECR_REPO_URI_PATTERN

    @property
    def repository_name(self) -> str:
        return self.get_match_groups()[2]

    @classmethod
    def from_components(  # type: ignore[override]
        cls,
        repository_name: str,
        account_id: str | None = None,
        region: str | None = None,
    ) -> "ECRRepositoryUri":
        """Generate a Repository URI.

        If account ID not provided, account Id of credentials is used.
        If region is not provided, region of credentials is used.

        Args:
            repository_name: Name of ECR repository.
            account_id: The registry ID. Defaults to None.
            region: AWS region. Defaults to None.

        Returns:
            AWS Repository URI.
        """
        registry_uri = ECRRegistryUri.from_components(account_id=account_id, region=region)
        return ECRRepositoryUri(f"{registry_uri}/{repository_name}")


class ECRImageUri(ECRRepositoryUri):
    regex_pattern = ECR_IMAGE_URI_PATTERN

    @property
    def image_tag(self) -> str | None:
        return self.get_match_groups()[-2]

    @property
    def image_digest(self) -> str | None:
        return self.get_match_groups()[-1]

    @classmethod
    def from_components(  # type: ignore[override]
        cls,
        repository_name: str,
        image_tag: str | None = None,
        image_digest: str | None = None,
        account_id: str | None = None,
        region: str | None = None,
    ) -> "ECRImageUri":
        """Generate an Image URI.

        If account ID not provided, account Id of credentials is used.
        If region is not provided, region of credentials is used.

        Args:
            repository_name: Name of ECR repository.
            image_tag: Tag associated with image. Defaults to None.
            image_digest: The image digest. Defaults to None.
            account_id: The registry ID. Defaults to None.
            region: AWS region. Defaults to None.

        Raises:
            ValueError: If both or neither image tag / image digest are provided.

        Returns:
            ECR Image URI.
        """
        if (image_tag and image_digest) or (not image_tag and not image_digest):
            raise ValueError(
                "Must provide EITHER image tag OR image digest. "
                f"image_tag={image_tag}, image_digest={image_digest}"
            )
        repo_uri = ECRRepositoryUri.from_components(
            repository_name=repository_name, account_id=account_id, region=region
        )
        image_id = f"{'@' if image_digest else ':'}{image_digest or image_tag}"
        return ECRImageUri(f"{repo_uri}{image_id}")


@dataclass
class ECRLogin:
    username: str
    password: str
    registry: str

    @property
    def auth_token(self) -> str:
        decoded_auth_token = f"{self.username}:{self.password}"
        return base64.b64encode(decoded_auth_token.encode()).decode()


class LifecyclePolicySelection(PydanticBaseModel):
    tag_status: Literal["tagged", "untagged", "any"]
    tag_prefix_list: list[str] | None
    count_type: Literal["imageCountMoreThan", "sinceImagePushed"]
    count_number: int
    count_unit: Literal["days"] | None = None

    @model_validator(mode="after")
    def _validate_fields(self):
        if self.count_type != "sinceImagePushed" and self.count_unit is not None:
            # https://docs.aws.amazon.com/AmazonECR/latest/userguide/LifecyclePolicies.html#lp_count_unit
            raise ValueError(f"Cannot specify 'countUnit' when countType={self.count_type}")
        if self.tag_status == "tagged" and not self.tag_prefix_list:
            # https://docs.aws.amazon.com/AmazonECR/latest/userguide/LifecyclePolicies.html#lp_tag_prefix_list
            raise ValueError(f"Must specify 'tagPrefixList' when tagStatus={self.tag_status}")
        return self

    def to_dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().to_dict(**kwargs)


class LifecyclePolicyAction(PydanticBaseModel):
    type: Literal["expire"] = "expire"

    def to_dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().to_dict(**kwargs)


class LifecyclePolicyRule(PydanticBaseModel):
    rule_priority: int = Field(serialization_alias="rulePriority", validation_alias="rulePriority")
    description: str
    selection: LifecyclePolicySelection
    action: LifecyclePolicyAction = Field(default_factory=LifecyclePolicyAction)

    @classmethod
    def REMOVE_UNTAGGED(cls, rule_priority: int = 1, days: int = 14) -> "LifecyclePolicyRule":
        return LifecyclePolicyRule(
            rule_priority=rule_priority,
            description=f"Remove untagged images after {days} days",
            selection=LifecyclePolicySelection(
                tag_status="untagged",
                tag_prefix_list=None,  # type: ignore[arg-type]
                count_type="sinceImagePushed",
                count_number=days,
                count_unit="days",
            ),
            action=LifecyclePolicyAction(type="expire"),
        )

    def to_dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().to_dict(**kwargs)


class LifecyclePolicy(PydanticBaseModel):
    rules: list[LifecyclePolicyRule]

    def model_post_init(self, __context):
        self.rules = self.reprioritize_rules(self.rules, in_place=True)

    @staticmethod
    def reprioritize_rules(
        rules: list[LifecyclePolicyRule], in_place: bool = False
    ) -> list[LifecyclePolicyRule]:
        rule_priority = 0

        sorted_rules: list[LifecyclePolicyRule]

        def sort_key(_):
            return _.rule_priority

        if in_place:
            rules.sort(key=sort_key)
            sorted_rules = rules
        else:
            sorted_rules = sorted(deepcopy(rules), key=sort_key)

        for rule in sorted_rules:
            if rule.rule_priority <= rule_priority:
                rule_priority += 1
                rule.rule_priority = rule_priority
            else:
                rule_priority = rule.rule_priority
        return sorted_rules

    @classmethod
    def from_rules(cls, *rules: LifecyclePolicyRule) -> "LifecyclePolicy":
        return LifecyclePolicy(rules=list(map(deepcopy, rules)))

    def to_dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        return super().to_dict(**kwargs)


class ResourceTag(TypedDict):
    Key: str
    Value: str


class TagMode(StrEnum):
    OVERWRITE = "overwrite"
    APPEND = "append"


T = TypeVar("T", bound="ECRResourceBase")


class ECRResourceBase(PydanticBaseModel):
    _client: ECRClient | None = PrivateAttr(default=None)
    _logger: logging.Logger | None = PrivateAttr(default=None)

    def __init__(self, client: ECRClient | None = None, **data: Any) -> None:
        super().__init__(**data)
        self._client = client

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.model_dump() == other.model_dump()

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.model_dump().items())))

    @property
    def client(self) -> ECRClient:
        if self._client is None:
            self._client = get_ecr_client(region=getattr(self, "region", None))
        return cast(ECRClient, self._client)

    @property
    def logger(self) -> logging.Logger:
        if self._logger is None:
            self._logger = logging.getLogger(f"{__name__}.{type(self).__name__}")
        return self._logger

    @property
    def uri(self) -> str:
        raise NotImplementedError("")  # pragma: no cover

    @classmethod
    def from_uri(cls: type[T], uri: str) -> T:
        raise NotImplementedError("")  # pragma: no cover


class ECRImage(ECRResourceBase):
    account_id: str
    region: str
    repository_name: str
    image_digest: str
    # https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest-field-descriptions
    _image_manifest: str | None = PrivateAttr(default=None)

    def __init__(
        self,
        *,
        account_id: str,
        region: str,
        repository_name: str,
        image_digest: str,
        image_manifest: str | None = None,
        client: ECRClient | None = None,
    ) -> None:
        super().__init__(  # type: ignore[call-arg]
            account_id=account_id,  # pyright: ignore[reportCallIssue]
            region=region,  # pyright: ignore[reportCallIssue]
            repository_name=repository_name,  # pyright: ignore[reportCallIssue]
            image_digest=image_digest,  # pyright: ignore[reportCallIssue]
            client=client,  # type: ignore[arg-type]
        )
        self._image_manifest = image_manifest

    @model_serializer(mode="wrap")
    def _serialize_with_manifest(self, handler) -> dict[str, Any]:
        d = handler(self)
        if self._image_manifest is not None:
            d["image_manifest"] = self._image_manifest
        return d

    @property
    def image_manifest(self) -> str:
        if self._image_manifest is None:
            self._image_manifest = self.fetch_image_manifest()
        return self._image_manifest

    @image_manifest.setter
    def image_manifest(self, value: str) -> None:
        self._image_manifest = value

    def fetch_image_manifest(self) -> str:
        response = self.client.batch_get_image(
            repositoryName=self.repository_name,
            registryId=self.account_id,
            imageIds=[ImageIdentifierTypeDef(imageDigest=self.image_digest)],
        )
        if len(response["images"]) == 0 or "imageManifest" not in response["images"][0]:
            raise ResourceNotFoundError(f"Could not resolve image manifest for {self.uri}")

        return response["images"][0]["imageManifest"]

    @property
    def image_pushed_at(self) -> datetime | None:
        return self.get_image_detail().get("imagePushedAt")

    @property
    def image_tags(self) -> list[str]:
        image_detail = self.get_image_detail()
        return image_detail.get("imageTags", [])

    @property
    def uri(self) -> str:
        return ECRImageUri.from_components(
            repository_name=self.repository_name,
            image_digest=self.image_digest,
            account_id=self.account_id,
            region=self.region,
        )

    def get_repository(self) -> "ECRRepository":
        return ECRRepository(
            account_id=self.account_id,
            region=self.region,
            repository_name=self.repository_name,
        )

    def get_image_detail(self) -> ImageDetailTypeDef:
        """Get image detail of this image from ECR.

        Returns:
            Image detail dictionary.
        """
        response = self.client.describe_images(
            repositoryName=self.repository_name,
            registryId=self.account_id,
            imageIds=[ImageIdentifierTypeDef(imageDigest=self.image_digest)],
        )
        image_details = response["imageDetails"]
        if len(image_details) == 0:
            raise ResourceNotFoundError(
                f"Could not resolve image detail for {self}"
            )  # pragma: no cover
        return image_details[0]

    def get_image_layers(self) -> list[LayerTypeDef]:
        """Get layers from image manifest into ECR Layer objects.

        The schema of the image manifest layers is defined here:
        https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest-field-descriptions

        Note:
            While docker image manifests can have multiple formats, ECR only supports
            the schema defined in the link above, a v2 single image manifest. There is
            a manifest list, that describes multiple architectures, but ECR does not support
            this. This method assumes the image manifest is in the correct format.

        Returns:
            List of ECR Image layers.
        """
        image_manifest = json.loads(self.image_manifest)

        return [
            LayerTypeDef(
                layerDigest=layer["digest"],
                layerAvailability="AVAILABLE",
                layerSize=layer["size"],
                mediaType=layer["mediaType"],
            )
            for layer in image_manifest["layers"]
        ]

    def get_image_config_layer(self) -> LayerTypeDef:
        """Get the image config layer from image manifest.

        The schema of the image manifest config layer is defined here:
        https://distribution.github.io/distribution/spec/manifest-v2-2/#image-manifest-field-descriptions

        Note:
            While docker image manifests can have multiple formats, ECR only supports
            the schema defined in the link above, a v2 single image manifest. There is
            a manifest list, that describes multiple architectures, but ECR does not support
            this. This method assumes the image manifest is in the correct format.

        Returns:
            Layer Type dict of the config object.
        """
        image_manifest = json.loads(self.image_manifest)
        layer = image_manifest["config"]
        return LayerTypeDef(
            layerDigest=layer["digest"],
            layerAvailability="AVAILABLE",
            layerSize=layer["size"],
            mediaType=layer["mediaType"],
        )

    def get_image_config(self) -> dict[str, Any]:
        """Get ECR or docker image configuration json metadata.

        Returns:
            Dictionary with image configuration.
        """
        # get image_manifest (sha256: hash)
        config_digest = json.loads(self.image_manifest)["config"]["digest"]
        registry = ECRRegistry(account_id=self.account_id, region=self.region)
        ecr_login = registry.get_ecr_login()

        response = requests.get(
            url=f"https://{registry.uri}/v2/{self.repository_name}/blobs/{config_digest}",
            headers={"Authorization": f"Basic {ecr_login.auth_token}"},
        )
        response.raise_for_status()
        return response.json()

    def add_image_tags(self, *image_tags: str):
        """Add tags to image.

        Args:
            *image_tags: Tags to add to image.
        """
        self.logger.info(f"Adding tags={image_tags} to {self.uri}")
        for tag in image_tags:
            self.put_image(image_tag=tag)

    def put_image(self, image_tag: str | None):
        """Make a call to put_image API to add image to ECR repository.

        This method will add an image to the ECR repository. If the image already exists,
        it will not raise an error. Instead, it will log that the image already
        exists with the given tag.

        Note:
            This operation does not push an image to the repository. It only adds
            the image manifest to the repository.

        Args:
            image_tag: Tag to associate with image. If None, image is untagged.
        """
        try:
            if image_tag is None:
                self.client.put_image(
                    registryId=self.account_id,
                    repositoryName=self.repository_name,
                    imageManifest=self.image_manifest,
                    imageDigest=self.image_digest,
                )
            else:
                self.client.put_image(
                    registryId=self.account_id,
                    repositoryName=self.repository_name,
                    imageManifest=self.image_manifest,
                    imageDigest=self.image_digest,
                    imageTag=image_tag,
                )
        except ClientError as e:
            # IF we receive a ImageAlreadyExistsException,
            # then we have nothing to worry about. Otherwise, raise error.
            if get_client_error_code(e) != "ImageAlreadyExistsException":
                raise e
            self.logger.info(f"Image already exists with tag={image_tag}")
        else:
            self.logger.info(f"Added new image with tag={image_tag}")

    @classmethod
    def from_repository_uri(cls, repository_uri: str, image_digest: str) -> "ECRImage":
        repo_uri = ECRRepositoryUri(repository_uri)
        return ECRImage.from_uri(f"{repo_uri}@{image_digest}")

    @classmethod
    def from_uri(cls, uri: str) -> "ECRImage":
        image_uri = ECRImageUri(uri)

        account_id = image_uri.account_id
        region = image_uri.region
        repo_name = image_uri.repository_name
        image_tag = image_uri.image_tag
        image_digest = image_uri.image_digest

        ecr = get_ecr_client(region)
        if image_digest is None:
            image_details = ecr.describe_images(
                repositoryName=repo_name,
                registryId=account_id,
                imageIds=[ImageIdentifierTypeDef(imageTag=image_tag)] if image_tag else [],
            )["imageDetails"][0]
            assert "imageDigest" in image_details
            image_digest = image_details["imageDigest"]

        return ECRImage(
            account_id=account_id,
            region=region,
            repository_name=repo_name,
            image_digest=image_digest,
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            self.account_id == other.account_id
            and self.region == other.region
            and self.repository_name == other.repository_name
            and self.image_digest == other.image_digest
        )

    def __hash__(self) -> int:
        return hash((self.account_id, self.region, self.repository_name, self.image_digest))

    def __repr__(self) -> str:
        return (
            f"ECRImage("
            f"account_id='{self.account_id}', "
            f"region='{self.region}', "
            f"repository_name='{self.repository_name}', "
            f"image_digest={self.image_digest[7:15]}..{self.image_digest[-4:]})"
        )


class ECRResource(ECRResourceBase):
    account_id: str
    region: str

    @property
    def arn(self) -> str:
        return f"arn:aws:ecr:{self.region}:{self.account_id}"

    def get_resource_tags(self) -> list[ResourceTag]:
        """Gets the tags for this ECR Resource.

        Returns:
            List of tags.
        """
        return [
            ResourceTag(Key=tag["Key"], Value=tag["Value"])
            for tag in self.client.list_tags_for_resource(resourceArn=self.arn)["tags"]
            if "Key" in tag and "Value" in tag
        ]

    def update_resource_tags(
        self, *tags: ResourceTag, mode: TagMode = cast(TagMode, TagMode.APPEND)
    ):
        """Updates the tags for an ECR Resource.

        An update can either append or overwrite the existing tags.

        Args:
            *tags: Resource tags to update.
            mode: Either append or overwrite tags of resource.
                Defaults to TagMode.APPEND.
        """
        tag_dict = {tag["Key"]: tag["Value"] for tag in tags}
        if mode == TagMode.OVERWRITE:
            existing_tags = self.get_resource_tags()
            tag_keys_to_remove = [
                existing_tag["Key"]
                for existing_tag in existing_tags
                if existing_tag["Key"] not in tag_dict
            ]
            self.client.untag_resource(resourceArn=self.arn, tagKeys=tag_keys_to_remove)
        self.client.tag_resource(
            resourceArn=self.arn,
            tags=[TagTypeDef(Key=key, Value=value) for key, value in tag_dict.items()],
        )


class ECRRepository(ECRResource):
    repository_name: str

    @property
    def arn(self) -> str:
        return f"{super().arn}:repository/{self.repository_name}"

    @property
    def uri(self) -> str:
        return ECRRepositoryUri.from_components(
            repository_name=self.repository_name,
            account_id=self.account_id,
            region=self.region,
        )

    def create(
        self,
        tags: list[ResourceTag] | None = None,
        image_tag_mutability: ImageTagMutabilityType = "MUTABLE",
        exists_ok: bool = True,
    ):
        """Create an ECR Repository.

        Args:
            tags: List of repo tags to add. Defaults to None.
            image_tag_mutability: Whether image tag is immutable. Defaults to "MUTABLE".
            exists_ok: Suppress error if repository already exists. Defaults to True.
        """
        if tags is None:
            tags = []

        try:
            self.client.create_repository(
                registryId=self.account_id,
                repositoryName=self.repository_name,
                tags=tags,
                imageTagMutability=image_tag_mutability,
            )
        except ClientError as e:
            # If the repo already exists, just move on.
            if get_client_error_code(e) == "RepositoryAlreadyExistsException" and exists_ok:
                # update tags
                if tags:
                    self.update_resource_tags(*tags)
            else:
                raise e

    def exists(self) -> bool:
        """Check if repository exists.

        Returns:
            True if repository exists.
        """
        try:
            self.client.describe_repositories(
                registryId=self.account_id, repositoryNames=[self.repository_name]
            )
        except ClientError as e:
            if get_client_error_code(e) == "RepositoryNotFoundException":
                return False
            else:
                raise e
        else:
            return True

    def put_lifecycle_policy(self, lifecycle_policy: LifecyclePolicy):
        self.client.put_lifecycle_policy(
            repositoryName=self.repository_name,
            lifecyclePolicyText=lifecycle_policy.to_json(sort_keys=True),
            registryId=self.account_id,
        )

    def delete(self, force: bool):
        """Delete the ECR repository described by this instance.

        Args:
            force: Ignore if images in repository.
        """
        self.client.delete_repository(
            registryId=self.account_id, repositoryName=self.repository_name, force=force
        )

    def get_image(
        self, image_tag: str | None = None, image_digest: str | None = None
    ) -> "ECRImage":
        """Get the image associated with the following tag or digest.

        Args:
            image_tag: Image tag. Defaults to None.
            image_digest: Image digest. Defaults to None.

        Returns:
            The image with the image tag or digest.
        """
        if (image_tag is None) == (image_digest is None):
            raise ValueError(
                f"Must provide image tag XOR digest. "
                f"provided image_tag={image_tag}, image_digest={image_digest}"
            )
        images = self.get_images()
        for image in images:
            if image_tag and image_tag in image.image_tags:
                return image
            elif image_digest and image_digest == image.image_digest:
                return image
        else:
            raise ResourceNotFoundError(
                f"Could not find an image in {self.uri} with tag={image_tag}"
            )

    def get_images(self, tag_status: TagStatusType = "ANY") -> list["ECRImage"]:
        """Fetches all images in a given repository.

        Args:
            tag_status: Filter non-tagged images. Defaults to "ANY".

        Returns:
            List of images in repository.
        """

        # To fetch necessary image info, we need to make two calls:
        #   1. Call the `list_images` API which will return imageDigest/imageTag.
        #       - This is a lightweight response which gives us digests
        #   2. Call the `batch_get_image` API which returns imageManifest info.
        #       - this is consolidated all in one go

        # Call 1: list_images
        paginator = self.client.get_paginator("list_images")
        list_image_request = ListImagesRequestPaginateTypeDef(
            repositoryName=self.repository_name,
            filter=ListImagesFilterTypeDef(tagStatus=tag_status),
        )
        image_digests: list[str] = sorted(
            list(
                {
                    image_id["imageDigest"]
                    for list_images_response in paginator.paginate(**list_image_request)
                    for image_id in list_images_response["imageIds"]
                    if "imageDigest" in image_id
                }
            )
        )
        if len(image_digests) == 0:
            return []

        # Call 2: batch_get_image
        response: BatchGetImageResponseTypeDef = self.client.batch_get_image(
            repositoryName=self.repository_name,
            registryId=self.account_id,
            imageIds=[ImageIdentifierTypeDef(imageDigest=digest) for digest in image_digests],
        )

        # Next we consolidate the results, ensuring that the image manifests
        # are all the same. If an image digest has differing manifests,
        # we should throw an error.
        digest_to_manifest_map: dict[str, str] = {}
        for image in response["images"]:
            image_digest = image["imageId"]["imageDigest"]  # type: ignore
            image_manifest = image["imageManifest"]  # type: ignore
            if image_digest in digest_to_manifest_map:
                if image_manifest != digest_to_manifest_map[image_digest]:
                    raise ValueError(
                        f"Not all image manifests are equivalent for {image_digest} in {self.uri}"
                    )
            else:
                digest_to_manifest_map[image_digest] = image_manifest

        return [
            ECRImage(
                account_id=self.account_id,
                region=self.region,
                repository_name=self.repository_name,
                image_digest=image_digest,
                image_manifest=image_manifest,
            )
            for image_digest, image_manifest in digest_to_manifest_map.items()
        ]

    @classmethod
    def from_uri(cls, uri: str) -> "ECRRepository":
        repo_uri = ECRRepositoryUri(uri)
        return ECRRepository(
            account_id=repo_uri.account_id,
            region=repo_uri.region,
            repository_name=repo_uri.repository_name,
        )

    @classmethod
    def from_arn(cls, arn: str) -> "ECRRepository":
        match = ECR_REPO_ARN_PATTERN.match(arn)
        if not match:
            raise InvalidAmazonResourceNameError(
                f"resource ARN = '{arn}' does not match {ECR_REPO_ARN_PATTERN} pattern"
            )
        return ECRRepository(
            account_id=match.group(2),
            region=match.group(1),
            repository_name=match.group(3),
        )

    @classmethod
    def from_name(
        cls,
        repository_name: str,
        account_id: str | None = None,
        region: str | None = None,
    ) -> "ECRRepository":
        region = get_region(region)
        account_id = account_id or get_account_id()
        return ECRRepository(account_id=account_id, region=region, repository_name=repository_name)


class ECRRegistry(ECRResource):
    @property
    def uri(self) -> str:
        return ECRRegistryUri.from_components(account_id=self.account_id, region=self.region)

    def get_repositories(
        self,
        repository_name: str | re.Pattern | None = None,
        repository_tags: list[ResourceTag] | None = None,
    ) -> list[ECRRepository]:
        """Filter repositories based on resource tags specified.

        Args:
            repository_name: Repository name or pattern to filter by.
            repository_tags: List of resource tags to filter by.

        Returns:
            Filtered list of repositories with resource tags.
        """
        repositories = self.list_repositories()
        filtered_repos = []
        for repo in repositories:
            if repository_tags:
                repo_tags = repo.get_resource_tags()
                if not all([filter_tag in repo_tags for filter_tag in repository_tags]):
                    continue
            if repository_name:
                if isinstance(repository_name, re.Pattern):
                    if not repository_name.match(repo.repository_name):
                        continue
                elif repository_name not in repo.repository_name:
                    continue
            filtered_repos.append(repo)
        return filtered_repos

    def list_repositories(self) -> list[ECRRepository]:
        """List all repositories in the Registry.

        Returns:
            List of repositories in the registry.
        """

        paginator = self.client.get_paginator("describe_repositories")
        repositories: list[ECRRepository] = []
        for describe_repos_response in paginator.paginate(registryId=self.account_id):
            for repository in describe_repos_response["repositories"]:
                assert "repositoryName" in repository
                repositories.append(
                    ECRRepository(
                        account_id=self.account_id,
                        region=self.region,
                        repository_name=repository["repositoryName"],
                    )
                )
        return repositories

    def get_ecr_login(self) -> ECRLogin:
        auth = self.client.get_authorization_token(registryIds=[self.account_id])
        if len(auth["authorizationData"]) != 1:
            raise AWSError(f"Could not resolve authorization token. Reponse: {auth}")
        auth_data_entry = auth["authorizationData"][0]
        auth_token = base64.b64decode(auth_data_entry["authorizationToken"]).decode()  # type: ignore
        username, password = auth_token.split(":")
        registry = auth_data_entry["proxyEndpoint"].replace("https://", "")  # type: ignore
        self.logger.debug(f"Registry (proxy endpoint): {registry}")
        assert "expiresAt" in auth_data_entry
        expires_at = auth_data_entry["expiresAt"].isoformat()
        self.logger.debug(f"Token expires at: {expires_at}")

        return ECRLogin(username=username, password=password, registry=registry)

    @classmethod
    def from_uri(cls, uri: str) -> "ECRRegistry":
        registry_uri = ECRRegistryUri(uri)
        return ECRRegistry(account_id=registry_uri.account_id, region=registry_uri.region)

    @classmethod
    def from_env(cls, region: str | None = None) -> "ECRRegistry":
        return ECRRegistry(account_id=get_account_id(), region=get_region(region))


def resolve_image_uri(name: str, default_tag: str | None = None) -> str:
    """Resolve full image URI from input name.

    Args:
        name: Partial or fully qualified uri, name of image or repository.
        default_tag: Default tag to use if not specified. Defaults to None.

    Returns:
        Fully qualified image URI.
    """

    try:
        uri = name

        if not ECRRegistryUri.is_prefixed(uri):
            uri = f"{ECRRegistryUri.from_components()}/{uri}"

        if ECRImageUri.is_valid(uri):
            return uri
        elif ECRRepositoryUri.is_valid(uri):
            repo = ECRRepository.from_uri(uri)

            if default_tag:
                image = repo.get_image(image_tag=default_tag)
                return image.uri
            else:
                # Fetch latest tagged
                def get_image_push_time(image: ECRImage) -> datetime:
                    if image.image_pushed_at is None:
                        raise RuntimeError(f"Couldn't get 'image_pushed_at' for: {image}")
                    return image.image_pushed_at

                images = sorted(repo.get_images("TAGGED"), key=get_image_push_time)
                return images[-1].uri
        else:
            raise ValueError(f"Could not resolve full URI for image {uri} (raw={name})")
    except Exception as e:
        msg = f"Couldn't resolve ECR image URI from {name} with error: {e}"
        logger.exception(msg)
        raise ResourceNotFoundError(msg) from e



import re
from typing import TYPE_CHECKING

import requests
from aibs_informatics_core.utils.logging import LoggingMixin
from botocore.exceptions import ClientError
from pydantic import Field

from aibs_informatics_aws_utils.core import get_client_error_code
from aibs_informatics_aws_utils.ecr.core import ECRImage, ECRRepository

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_ecr import ECRClient
    from mypy_boto3_ecr.type_defs import LayerTypeDef
else:
    ECRClient = object
    LayerTypeDef = dict

from aibs_informatics_core.models.base import PydanticBaseModel


class ReplicateImageRequest(PydanticBaseModel):
    source_image: ECRImage
    destination_repository: ECRRepository
    destination_image_tags: list[str] = Field(default_factory=list)


class ReplicateImageResponse(PydanticBaseModel):
    destination_image: ECRImage


class ECRImageReplicator(LoggingMixin):
    def process_request(self, request: ReplicateImageRequest) -> ReplicateImageResponse:
        self.log.info(f"processing request='{request}")
        dest_image = self.replicate(
            source_image=request.source_image,
            destination_repository=request.destination_repository,
            destination_image_tags=request.destination_image_tags,
        )
        self.log.info(f"Replicated image to repo with uri={dest_image.uri}")
        return ReplicateImageResponse(destination_image=dest_image)

    def replicate(
        self,
        source_image: ECRImage,
        destination_repository: ECRRepository,
        destination_image_tags: list[str] | None = None,
    ) -> ECRImage:
        """Copies the source image into the destination repository.

        This allows the user to facilitate replication:

        - **between AWS accounts**
        - **to a new repository**
        - **with new tags**

        Note:
            This has exactly the same effect as `docker pull; docker tag; docker push`,
            but is done with the AWS ECR SDK so we can run this in a lambda when our
            stack updates.

        Args:
            source_image: Configuration of image to pull.
            destination_repository: Repository to push to.
            destination_image_tags: Optional list of tags to apply to the destination image.

        Returns:
            Configuration of destination image pushed.
        """
        self.log.info(
            f"Starting Image Replication "
            f"[source={source_image.uri}, destination={destination_repository.uri}] "
            f"with tags={destination_image_tags}"
        )

        self.upload_layers(
            source_image=source_image, destination_repository=destination_repository
        )

        self.logger.info(f"Putting image to {destination_repository.uri}")

        self.put_image(
            source_image=source_image,
            destination_repository=destination_repository,
            destination_image_tags=destination_image_tags,
        )
        self.logger.info(f"Completed putting image to {destination_repository.uri}.")

        return ECRImage(
            account_id=destination_repository.account_id,
            region=destination_repository.region,
            repository_name=destination_repository.repository_name,
            image_digest=source_image.image_digest,
            client=destination_repository.client,
        )

    def upload_layers(self, source_image: ECRImage, destination_repository: ECRRepository):
        """Upload image layers of the source image to the destination repository.

        Args:
            source_image: Source image that has been copied.
            destination_repository: Destination repository.
        """
        self.logger.info(
            f"Uploading layers from {source_image.uri} to {destination_repository.uri}."
        )

        layers = source_image.get_image_layers()
        config_layer = source_image.get_image_config_layer()

        all_layers = layers + [config_layer]

        self.logger.info(
            f"Source image {source_image.uri} has {len(layers)} layers and "
            f"1 config layer totalling {len(all_layers)} layers"
        )

        self._upload_layers(
            source_repository=source_image.get_repository(),
            destination_repository=destination_repository,
            layers=all_layers,
            check_if_exists=True,
        )

    def put_image(
        self,
        source_image: ECRImage,
        destination_repository: ECRRepository,
        destination_image_tags: list[str] | None = None,
    ) -> ECRImage:
        """Put image manifest and tags for an image.

        This is the final step in copying an ECR image. It must be done
        once all the layers of the image have been uploaded.

        If the put_image request fails with a 'LayersNotFoundException',
        this will attempt to upload the missing layers and retry
        the put_image request.

        Args:
            source_image: Source image that has been copied.
            destination_repository: Destination repository.
            destination_image_tags: Destination image tags. Optional.
                If not provided, source image's tags are added.

        Returns:
            Destination ECR Image.
        """
        tags = (
            source_image.image_tags if destination_image_tags is None else destination_image_tags
        )

        self.logger.info(f"putting image to {destination_repository.uri} with {tags} tags")
        dest_image = ECRImage(
            account_id=destination_repository.account_id,
            region=destination_repository.region,
            repository_name=destination_repository.repository_name,
            image_digest=source_image.image_digest,
            image_manifest=source_image.image_manifest,
            client=destination_repository.client,
        )
        try:
            dest_image.put_image(None)
            if tags:
                dest_image.add_image_tags(*tags)

        except ClientError as e:
            # IF we have a LayersNotFoundException, then we will retry to upload layers
            if get_client_error_code(e) != "LayersNotFoundException":
                self.log_stacktrace(
                    f"Error putting image to {destination_repository.uri} with {tags}",
                    e,
                )
                raise e
            self.logger.warning(
                f"Received 'LayerNotFoundException' while putting image. "
                f"Resolving missing layers of {source_image} specified in {e}"
            )
            missing_layers = self._get_missing_layers(
                client=source_image.client,
                repository_name=source_image.repository_name,
                put_image_error=e,
            )
            self.logger.info(f"Identified {len(missing_layers)} missing layers. Uploading layers")
            self._upload_layers(
                source_repository=source_image.get_repository(),
                destination_repository=destination_repository,
                layers=missing_layers,
                check_if_exists=False,
            )

            self.logger.info("uploaded all missing layers, retrying put_image request")
            self.put_image(source_image, destination_repository, destination_image_tags)
        return dest_image

    def _upload_layers(
        self,
        source_repository: ECRRepository,
        destination_repository: ECRRepository,
        layers: list[LayerTypeDef],
        check_if_exists: bool = True,
    ):
        """Upload image layers of the source image to the destination repository.

        Args:
            source_repository: Source repository of layers.
            destination_repository: Destination repository.
            layers: Layers from source repository.
            check_if_exists: Whether to check if layer already exists before uploading.
                Defaults to True.
        """
        for i, layer in enumerate(layers):
            self.logger.info(f"Starting upload of layer {i + 1} / {len(layers)}")
            if check_if_exists:
                layer_exists = self._check_layer_exists(
                    client=destination_repository.client,
                    repository_name=destination_repository.repository_name,
                    layer_digest=layer.get("layerDigest", ""),
                )
                if layer_exists:
                    self.logger.info(
                        f"layer {layer} already exists in {destination_repository.uri}"
                    )
                    continue

            self._upload_layer(
                source_repository=source_repository,
                destination_repository=destination_repository,
                layer=layer,
            )

    def _upload_layer(
        self,
        source_repository: ECRRepository,
        destination_repository: ECRRepository,
        layer: LayerTypeDef,
    ):
        self.logger.info(
            f"uploading layer {layer} from {source_repository.uri} to {destination_repository.uri}"
        )
        layer_digest = layer["layerDigest"]  # type: ignore
        layer_size = layer["layerSize"]  # type: ignore

        self.logger.info(f"getting private download url for layer {layer_digest}")
        download_url = self._get_download_url_for_layer(
            client=source_repository.client,
            repository_name=source_repository.repository_name,
            layer_digest=layer_digest,
        )

        self.logger.info(f"initiating upload to destination {destination_repository.uri}")
        initiate_layer_upload_response = destination_repository.client.initiate_layer_upload(
            registryId=destination_repository.account_id,
            repositoryName=destination_repository.repository_name,
        )

        upload_id = initiate_layer_upload_response["uploadId"]
        part_size = initiate_layer_upload_response["partSize"]

        current_source_layer_size = layer_size
        transfer_bytes_remaining = current_source_layer_size
        self.logger.info(
            f"uploading all {current_source_layer_size:,} bytes of layer {layer_digest} "
            f"in chunks of {part_size:,} with upload id {upload_id}"
        )

        while transfer_bytes_remaining > 0:
            part_first_byte = current_source_layer_size - transfer_bytes_remaining
            part_last_byte = min(current_source_layer_size, part_first_byte + part_size) - 1

            self.logger.info(
                f"uploading bytes {part_first_byte:,} - {part_last_byte:,} of layer {layer_digest}"
            )
            last_byte_received = self._upload_layer_part(
                client=destination_repository.client,
                repository_name=destination_repository.repository_name,
                download_url=download_url,
                upload_id=upload_id,
                part_first_byte=part_first_byte,
                part_last_byte=part_last_byte,
            )
            transfer_bytes_remaining = current_source_layer_size - last_byte_received - 1
            self.logger.info(
                f"successfully uploaded bytes {part_first_byte:,} to {last_byte_received:,} "
                f"completed {last_byte_received / layer_size:.2%}"
            )

        self.logger.info(
            f"Successfully uploaded all layer parts for {layer_digest}. Completing layer upload"
        )
        self._complete_layer_upload(
            client=destination_repository.client,
            repository_name=destination_repository.repository_name,
            upload_id=upload_id,
            layer_digest=layer_digest,
        )

    # ------------------------------------------------------------------------
    # Wrapper methods for layer related API calls
    # ------------------------------------------------------------------------

    def _check_layer_exists(
        self, client: ECRClient, repository_name: str, layer_digest: str
    ) -> bool:
        """
        Checks if a layer exists in the provided ECR Repository
        """
        self.logger.info(f"checking if layer {layer_digest} exists in {repository_name}")
        response = client.batch_check_layer_availability(
            repositoryName=repository_name, layerDigests=[layer_digest]
        )
        layers = response["layers"]
        if len(layers) != 1:
            raise AttributeError(
                f"size of layers list from batch_check_layer_availability"
                f"response {response} must equal 1 "
                f"repository={repository_name}, "
                f"and layerDigest={layer_digest}"
            )

        return response["layers"][0].get("layerAvailability") == "AVAILABLE"

    def _get_download_url_for_layer(
        self, client: ECRClient, repository_name: str, layer_digest: str
    ) -> str:
        """Gets authenticated download_url for the provided layer"""

        get_download_url_for_layer_response = client.get_download_url_for_layer(
            repositoryName=repository_name, layerDigest=layer_digest
        )
        return get_download_url_for_layer_response["downloadUrl"]

    def _upload_layer_part(
        self,
        client: ECRClient,
        repository_name: str,
        download_url: str,
        upload_id: str,
        part_first_byte: int,
        part_last_byte: int,
    ) -> int:
        """Uploads the range of bytes from the provided download url to destination repository.

        Args:
            client: ECR client.
            repository_name: Destination repository to which layer part uploaded.
            download_url: S3 url for the layer from `_get_download_url_for_layer`.
            upload_id: Upload id from `initiate_layer_upload`.
            part_first_byte: First byte to upload.
            part_last_byte: Last byte to upload.

        Returns:
            The last byte received from the `upload_layer_part` request
            (not necessarily the part_last_byte).
        """

        http_response = requests.request(
            "GET",
            download_url,
            headers={"Range": f"bytes={part_first_byte}-{part_last_byte}"},
            stream=False,
        )

        layer_part_bytes = b""
        for b in http_response.iter_content(chunk_size=None):
            layer_part_bytes += b

        upload_layer_part_response = client.upload_layer_part(
            repositoryName=repository_name,
            uploadId=upload_id,
            partFirstByte=part_first_byte,
            partLastByte=part_last_byte,
            layerPartBlob=layer_part_bytes,
        )

        return upload_layer_part_response["lastByteReceived"]

    def _complete_layer_upload(
        self, client: ECRClient, repository_name: str, upload_id: str, layer_digest: str
    ):
        """Tells ECR that layer upload has completed.

        If the layer already exists in the AWS account / repository
        will return with success.

        Args:
            client: ECR client.
            repository_name: Destination repository to which layer part uploaded.
            upload_id: Upload id for the layer from `initiate_layer_upload`.
            layer_digest: SHA of the layer.
        """

        try:
            complete_layer_upload_response = client.complete_layer_upload(
                repositoryName=repository_name,
                uploadId=upload_id,
                layerDigests=[layer_digest],
            )
            assert complete_layer_upload_response["layerDigest"] == layer_digest

        except ClientError as e:
            if get_client_error_code(e) != "LayerAlreadyExistsException":
                self.logger.error(
                    f"Unexpected Error completing layer upload "
                    f"with upload id: {upload_id} "
                    f"for repository: {repository_name} "
                )
                raise e

            self.logger.info(
                f"Layer with digest {layer_digest} exits in repository: {repository_name}. "
                "returning with success."
            )

    def _get_missing_layers(
        self, client: ECRClient, repository_name: str, put_image_error: ClientError
    ) -> list[LayerTypeDef]:
        """Gets missing layers from a ClientError while putting image.

        Args:
            client: ECR client.
            repository_name: Source repository containing layer of layer digest.
            put_image_error: Error thrown after putting image.

        Returns:
            List of missing layers.
        """

        error_message = put_image_error.response.get("Error", {}).get("Message", "Unknown")

        match = re.search(r".*\[(sha.*)\].*", error_message)
        if not match:
            raise ValueError(f"unable to extract missing shas from error message: {error_message}")
        missing_layer_digests = [item.strip() for item in match.group(1).split(",")]
        return [
            self._get_layer_from_digest(
                client=client,
                repository_name=repository_name,
                layer_digest=missing_layer_digest,
            )
            for missing_layer_digest in missing_layer_digests
        ]

    def _get_layer_from_digest(
        self, client: ECRClient, repository_name: str, layer_digest: str
    ) -> LayerTypeDef:
        """Get the layer info of an image's layer.

        Args:
            client: ECR client.
            repository_name: Source repository containing layer of layer digest.
            layer_digest: Layer sha256 identifier.

        Returns:
            Image layer digest and size in bytes.
        """
        response = client.batch_check_layer_availability(
            repositoryName=repository_name, layerDigests=[layer_digest]
        )
        layers = response["layers"]
        if len(layers) != 1:
            raise AttributeError(
                f"size of layers list from batch_check_layer_availability"
                f"response {response} must equal 1 "
                f"for repository {repository_name} and layerDigest {layer_digest}"
            )
        if layers[0].get("layerAvailability") == "UNAVAILABLE":
            raise AttributeError(
                f"Layer with digest {layer_digest} from repository {repository_name} "
                f"is unavailable! Unable to extract layer size."
            )
        return layers[0]



__all__ = [
    "get_account_id",
    "get_client",
    "get_client_error_code",
    "get_client_error_message",
    "get_region",
    "get_resource",
    "get_session",
    "AWSService",
    "AWS_REGION_VAR",
]
import logging
import os
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, Literal, TypeVar, cast

import boto3
from aibs_informatics_core.models.aws.core import AWSRegion
from aibs_informatics_core.models.aws.iam import IAMArn, UserId
from aibs_informatics_core.utils.decorators import cache
from boto3 import Session
from boto3.resources.base import ServiceResource
from botocore.client import BaseClient, ClientError
from botocore.config import Config
from botocore.session import Session as BotocoreSession

if TYPE_CHECKING:  # pragma: no cover
    from mypy_boto3_apigateway import APIGatewayClient
    from mypy_boto3_athena import AthenaClient
    from mypy_boto3_batch import BatchClient
    from mypy_boto3_dynamodb import DynamoDBClient, DynamoDBServiceResource
    from mypy_boto3_ec2 import EC2Client, EC2ServiceResource
    from mypy_boto3_ecr import ECRClient
    from mypy_boto3_ecs import ECSClient
    from mypy_boto3_efs import EFSClient
    from mypy_boto3_fsx import FSxClient
    from mypy_boto3_lambda import LambdaClient
    from mypy_boto3_logs import CloudWatchLogsClient
    from mypy_boto3_s3 import S3Client, S3ServiceResource
    from mypy_boto3_secretsmanager import SecretsManagerClient
    from mypy_boto3_ses import SESClient
    from mypy_boto3_sns import SNSClient, SNSServiceResource
    from mypy_boto3_sqs import SQSClient, SQSServiceResource
    from mypy_boto3_ssm import SSMClient
    from mypy_boto3_stepfunctions import SFNClient
    from mypy_boto3_sts import STSClient
    from mypy_boto3_sts.type_defs import GetCallerIdentityResponseTypeDef
else:
    AthenaClient = object
    APIGatewayClient = object
    BatchClient = object
    CloudWatchLogsClient = object
    DynamoDBClient, DynamoDBServiceResource = object, object
    EC2Client, EC2ServiceResource = object, object
    ECRClient = object
    ECSClient = object
    EFSClient = object
    FSxClient = object
    GetCallerIdentityResponseTypeDef = dict
    LambdaClient = object
    S3Client, S3ServiceResource = object, object
    SecretsManagerClient = object
    SESClient = object
    SNSClient, SNSServiceResource = object, object
    SQSClient, SQSServiceResource = object, object
    SSMClient = object
    STSClient = object
    SFNClient = object


from aibs_informatics_aws_utils.exceptions import AWSError

logger = logging.getLogger(__name__)  # type: logging.Logger


# ----------------------------------------------------------------------------
# AWS Session / Account / Region utilties
# ----------------------------------------------------------------------------


AWS_REGION_VAR = "AWS_REGION"


def get_session(session: Session | BotocoreSession | None = None) -> Session:
    """Get or create a boto3 Session.

    Args:
        session (Optional[Union[Session, BotocoreSession]]): An existing Session
            or BotocoreSession to use. If None, creates a new default Session.

    Returns:
        A boto3 Session object.
    """
    if not session:
        return Session()
    elif isinstance(session, BotocoreSession):
        return Session(botocore_session=session)
    else:
        return session


def get_region(region: str | None = None) -> str:
    """Get and sanitize region value

    Will retrieve the current region from a newly created boto3 session.
    This is helpful for getting the region dynamically within a Lambda
    function or from the environment if that is not available.

    Logic of priority for resolving region:
        * User input
        * boto3.Session
        * ENV VAR "AWS_REGION"
        * ENV VAR "REGION"

    Subsequently, the region value is validated against a regex

    Args:
        region (str, optional): user provided region. Defaults to None.

    Raises:
        ApplicationException:
            - If region cannot be resolved
            - If region is not correct format

    Returns:
        AWS Region
    """

    # If not provided, check session
    if not region:
        session = Session()
        region = session.region_name
    # If session does not resolve, check environment variables
    if not region:
        # Attempt to get it from an environment variable.
        ENV_VARS = ["AWS_REGION", "REGION"]
        logger.debug(
            "Could not resolve region from session. "
            f"Looking for following env variables: {ENV_VARS}"
        )
        region = next((os.environ.get(key) for key in ENV_VARS if os.environ.get(key)), None)

    if not region:
        error_msg = "Could not determine region from default session or environment"
        logger.error(error_msg)
        raise AWSError(error_msg)
    try:
        region = AWSRegion(region)
        assert region is not None  # mollify mypy
    except Exception as e:
        raise AWSError from e
    return region


def get_account_id() -> str:
    """Get the AWS account ID from the current credentials.

    Returns:
        The AWS account ID as a string.
    """
    return get_caller_identity()["Account"]


def get_user_id() -> UserId:
    """Get the IAM user ID from the current credentials.

    Returns:
        The IAM user ID.
    """
    return UserId(get_caller_identity()["UserId"])


def get_iam_arn() -> IAMArn:
    """Get the IAM ARN from the current credentials.

    Returns:
        The IAM ARN of the current identity.
    """
    return IAMArn(get_caller_identity()["Arn"])


def get_caller_identity() -> GetCallerIdentityResponseTypeDef:
    """Get the caller identity from AWS STS.

    Returns:
        The caller identity response containing Account, Arn, and UserId.
    """
    return boto3.client("sts").get_caller_identity()


# ----------------------------------------------------------------------------
# Common utilities
# ----------------------------------------------------------------------------


def get_client_error_code(client_error: ClientError) -> str:
    """Extract the error code from a boto3 ClientError.

    Args:
        client_error (ClientError): The ClientError exception.

    Returns:
        The error code string, or "Unknown" if not found.
    """
    return client_error.response.get("Error", {}).get("Code", "Unknown")


def get_client_error_message(client_error: ClientError) -> str:
    """Extract the error message from a boto3 ClientError.

    Args:
        client_error (ClientError): The ClientError exception.

    Returns:
        The error message string, or "Unknown" if not found.
    """
    return client_error.response.get("Error", {}).get("Message", "Unknown")


def client_error_code_check(client_error: ClientError, *error_codes: str) -> bool:
    """Check if a ClientError matches any of the specified error codes.

    Args:
        client_error (ClientError): The ClientError exception to check.
        *error_codes (str): One or more error codes to check against.

    Returns:
        True if the error code matches any of the specified codes.
    """
    return get_client_error_code(client_error) in error_codes


# ----------------------------------------------------------------------------
# AWS Client / Resource Utilities
# ----------------------------------------------------------------------------

Resources = Literal[
    "dynamodb",
    "ec2",
    "s3",
    "sns",
    "sqs",
]
Clients = Literal[
    "athena",
    "apigateway",
    "batch",
    "dynamodb",
    "ec2",
    "ecr",
    "ecs",
    "efs",
    "fsx",
    "lambda",
    "logs",
    "s3",
    "secretsmanager",
    "ses",
    "sns",
    "sqs",
    "ssm",
    "sts",
    "stepfunctions",
]
Services = Literal[Clients, Resources]


ClientType = TypeVar("ClientType", bound=BaseClient)
ResourceType = TypeVar("ResourceType", bound=ServiceResource)


@cache
def get_client(
    service: Clients,
    session: boto3.Session | None = None,
    region: str | None = None,
    **kwargs,
):
    """Get a boto3 client object
    Notes:

    - If session is not provided, then the default session is used
    - Order of priority for region is
      1. region parameter
      2. "region_name" in **kwargs
      3. fallback to whatever is fetched in get_region

    Args:
        service (Clients): The intended service to build client for
        session (boto3.Session, optional): A customized session object. Defaults to None.
        region (Optional[str], optional): An explicit region. Defaults to None.

    Returns:
        botocore.client.BaseClient: A boto3 Client object
    """
    region_name = get_region(region=region or kwargs.get("region_name"))
    if region_name:
        kwargs["region_name"] = region_name

    # If config for our client is not set, we want to set it to use "standard" mode
    # (default is "legacy") and increase the number of retries to 5 (default is 3)
    # See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html#available-retry-modes
    config: Config | None = kwargs.pop("config", None)
    default_config = Config(
        connect_timeout=120, read_timeout=120, retries={"max_attempts": 6, "mode": "standard"}
    )
    if config is None:
        config = default_config
    else:
        # Have values in pre-existing config (if it exists) take precedence over default_config
        config = default_config.merge(other_config=config)

    session = session or boto3.Session()
    return session.client(service, config=config, **kwargs)


@cache
def get_resource(
    service: Resources,
    session: boto3.Session | None = None,
    region: str | None = None,
    **kwargs,
) -> ServiceResource:
    """Get a boto3 resource object
    Notes:

    - If session is not provided, then the default session is used
    - Order of priority for region is
      1. region parameter
      2. "region_name" in **kwargs
      3. fallback to whatever is fetched in get_region

    Args:
        service (Resources): The intended service to build resource from
        session (boto3.Session, optional): A customized session object. Defaults to None.
        region (Optional[str], optional): An explicit region. Defaults to None.

    Returns:
        boto3.resources.base.ServiceResource: A ServiceResource object
    """
    region_name = get_region(region=region or kwargs.get("region_name"))
    if region_name:
        kwargs["region_name"] = region_name

    # If config for our client is not set, we want to set it to use "standard" mode
    # (default is "legacy") and increase the number of retries to 5 (default is 3)
    # See: https://boto3.amazonaws.com/v1/documentation/api/latest/guide/retries.html#available-retry-modes
    config: Config | None = kwargs.pop("config", None)
    default_config = Config(
        connect_timeout=120, read_timeout=120, retries={"max_attempts": 6, "mode": "standard"}
    )
    if config is None:
        config = default_config
    else:
        # Have values in pre-existing config (if it exists) take precedence over default_config
        config = default_config.merge(other_config=config)

    session = session or boto3.Session()
    return session.resource(service, config=config, **kwargs)


@dataclass
class AWSServiceProvider(Generic[ClientType]):
    """Provider for creating typed AWS service clients.

    Attributes:
        service_name: The name of the AWS service.
    """

    service_name: Services

    def get_client(self, region: str | None = None, **kwargs) -> ClientType:
        """Get a typed client for this AWS service.

        Args:
            region (Optional[str]): AWS region. Defaults to None (uses default region).
            **kwargs: Additional arguments passed to boto3 client creation.

        Returns:
            A typed boto3 client for the service.
        """
        return cast(ClientType, get_client(self.service_name, region=region, **kwargs))


@dataclass
class AWSServiceAndResourceProvider(
    AWSServiceProvider[ClientType], Generic[ClientType, ResourceType]
):
    """Provider for creating typed AWS service clients and resources.

    Extends AWSServiceProvider to also support resource-based access patterns.

    Attributes:
        service_name: The name of the AWS service (must support resources).
    """

    service_name: Resources

    def get_resource(self, region: str | None = None, **kwargs) -> ResourceType:
        """Get a typed resource for this AWS service.

        Args:
            region (Optional[str]): AWS region. Defaults to None (uses default region).
            **kwargs: Additional arguments passed to boto3 resource creation.

        Returns:
            A typed boto3 resource for the service.
        """
        return cast(ResourceType, get_resource(self.service_name, region=region, **kwargs))


class AWSService:
    ATHENA = AWSServiceProvider[AthenaClient]("athena")
    API_GATEWAY = AWSServiceProvider[APIGatewayClient]("apigateway")
    BATCH = AWSServiceProvider[BatchClient]("batch")
    DYNAMO_DB = AWSServiceAndResourceProvider[DynamoDBClient, DynamoDBServiceResource]("dynamodb")
    EC2 = AWSServiceAndResourceProvider[EC2Client, EC2ServiceResource]("ec2")
    ECR = AWSServiceProvider[ECRClient]("ecr")
    ECS = AWSServiceProvider[ECSClient]("ecs")
    EFS = AWSServiceProvider[EFSClient]("efs")
    FSX = AWSServiceProvider[FSxClient]("fsx")
    LAMBDA = AWSServiceProvider[LambdaClient]("lambda")
    LOGS = AWSServiceProvider[CloudWatchLogsClient]("logs")
    S3 = AWSServiceAndResourceProvider[S3Client, S3ServiceResource]("s3")
    SECRETSMANAGER = AWSServiceProvider[SecretsManagerClient]("secretsmanager")
    SES = AWSServiceProvider[SESClient]("ses")
    SNS = AWSServiceAndResourceProvider[SNSClient, SNSServiceResource]("sns")
    SQS = AWSServiceAndResourceProvider[SQSClient, SQSServiceResource]("sqs")
    SSM = AWSServiceProvider[SSMClient]("ssm")
    STS = AWSServiceProvider[STSClient]("sts")
    STEPFUNCTIONS = AWSServiceProvider[SFNClient]("stepfunctions")



from botocore.auth import SigV4Auth
from botocore.awsrequest import AWSRequest
from botocore.compat import parse_qsl, urlparse
from botocore.session import Session
from requests.auth import AuthBase

from aibs_informatics_aws_utils.core import get_session


class IamAWSRequestsAuth(AuthBase):
    """IAM authorizer for signing HTTP requests with AWS SigV4.

    This class can be used with the `requests` library to automatically sign
    HTTP requests using AWS IAM credentials.

    Args:
        session (Optional[Session]): Optional botocore Session object for credentials.
        service_name (str): The AWS service name for signing. Defaults to "execute-api".

    Example:
        ```python
        auth = IamAWSRequestsAuth()
        response = requests.get(url, auth=auth)

        # With custom session and service
        auth = IamAWSRequestsAuth(boto3.Session(), 'execute-api')
        ```
    """

    def __init__(self, session: Session | None = None, service_name: str = "execute-api"):
        self.boto3_session = get_session(session)
        credentials = self.boto3_session.get_credentials()
        if not credentials:
            raise ValueError("No AWS credentials found")

        self.sigv4 = SigV4Auth(
            credentials=credentials.get_frozen_credentials(),
            service_name=service_name,
            region_name=self.boto3_session.region_name,
        )

    def __call__(self, request):
        # Parse request URL
        url = urlparse(request.url)

        # Prepare AWS request
        awsrequest = AWSRequest(
            method=request.method,
            url=f"{url.scheme}://{url.netloc}{url.path}",
            data=request.body if hasattr(request, "body") else (request.json or request.data),
            params=dict(parse_qsl(url.query)),
        )

        # Sign request
        self.sigv4.add_auth(awsrequest)

        # Re-add original headers
        for key, val in request.headers.items():
            if key not in awsrequest.headers:
                awsrequest.headers[key] = val

        # Return prepared request
        return awsrequest.prepare()
